# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Base for HTML controls."""

import abc
import contextlib
import inspect
import sys
from typing import Annotated, Any, Dict, Iterator, List, Optional, Union

from pyglove.core import object_utils
from pyglove.core.symbolic import object as pg_object
from pyglove.core.views.html import base

Html = base.Html


try:
  _notebook = sys.modules['IPython'].display
except Exception:  # pylint: disable=broad-except
  _notebook = None


class HtmlControl(pg_object.Object):
  """Base class for HTML controls."""

  id: Annotated[
      Optional[str],
      'Optional ID of the root element.'
  ] = None

  css_classes: Annotated[
      List[str],
      'Optional CSS classes of the root element.'
  ] = []

  styles: Annotated[
      Dict[str, Any],
      'Ad-hoc styles (vs. class-based) of the root element.'
  ] = {}

  interactive: Annotated[
      bool,
      (
          'Whether the control is interactive. If False, optimizations will be '
          'applied to reduce the size of the HTML.'
      )
  ] = False

  def _on_bound(self):
    super()._on_bound()
    self._rendered = False
    self._css_styles = []
    self._scripts = []

  def add_style(self, *css: str) -> 'HtmlControl':
    """Adds CSS styles to the HTML."""
    self._css_styles.extend(css)
    return self

  def add_script(self, *scripts: str) -> 'HtmlControl':
    self._scripts.extend(scripts)
    return self

  def to_html(self, **kwargs) -> Html:
    """Returns the HTML representation of the control."""
    self._rendered = True
    html = self._to_html(**kwargs)
    return html.add_style(*self._css_styles).add_script(*self._scripts)

  @abc.abstractmethod
  def _to_html(self, **kwargs) -> Html:
    """Returns the HTML representation of the control."""

  @classmethod
  @contextlib.contextmanager
  def track_scripts(cls) -> Iterator[List[str]]:
    del cls
    all_tracked = object_utils.thread_local_get(_TLS_TRACKED_SCRIPTS, [])
    current = []
    all_tracked.append(current)
    object_utils.thread_local_set(_TLS_TRACKED_SCRIPTS, all_tracked)
    try:
      yield current
    finally:
      all_tracked.pop(-1)
      if not all_tracked:
        object_utils.thread_local_del(_TLS_TRACKED_SCRIPTS)

  def _sync_members(self, **fields) -> None:
    """Synchronizes displayed values to members."""
    self.rebind(fields, skip_notification=True, raise_on_no_change=False)

  def _run_javascript(self, code: str) -> None:
    """Runs the given JavaScript code."""
    if not self.interactive:
      raise ValueError(
          f'Non-interactive control {self} cannot be updated. '
          'Please set `interactive=True` in the constructor.'
      )
    if not self._rendered:
      return

    code = inspect.cleandoc(code)
    if _notebook is not None:
      _notebook.display(_notebook.Javascript(code))

    # Track script execution.
    all_tracked = object_utils.thread_local_get(_TLS_TRACKED_SCRIPTS, [])
    for tracked in all_tracked:
      tracked.append(code)

  def element_id(self, child: Optional[str] = None) -> Optional[str]:
    """Returns the element id of this control or a child."""
    if self.id is not None:
      return self.id
    elif not self.interactive:
      return None
    elif child is None:
      return f'control-{id(self)}'
    else:
      return f'control-{id(self)}-{child}'

  def _update_content(
      self,
      content: Union[str, Html],
      child: Optional[str] = None,
  ) -> None:
    """Updates the content of the control."""
    if isinstance(content, str):
      return self._update_text(content, child)
    else:
      return self._update_inner_html(content, child)

  def _update_text(
      self,
      content: str,
      child: Optional[str] = None
  ) -> str:
    """Updates the content of the control."""
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.textContent = "{Html.escape(content, javascript_str=True)}";
        """
    )
    return content

  def _update_inner_html(
      self,
      html: base.Html,
      child: Optional[str] = None,
  ) -> base.Html:
    """Updates the inner HTML of the control."""
    js_html = Html.escape(html, javascript_str=True)
    assert isinstance(js_html, Html), js_html
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.innerHTML = "{js_html.to_str(content_only=True)}";
        """
    )
    return html

  def _update_style(
      self,
      styles: Dict[str, Any],
      *,
      child: Optional[str] = None,
      updates_only: bool = False,
  ) -> Dict[str, Any]:
    """Updates the style of the control."""
    updated_styles = {} if updates_only else dict(self.styles or {})
    updated_styles.update(styles)
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.style = "{Html.style_str(updated_styles)}";
        """
    )
    return updated_styles

  def _update_property(
      self,
      name: str,
      value: str,
      child: Optional[str] = None,
  ) -> str:
    """Updates a property of the control."""
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.{name} = "{value}";
        """
    )
    return value

  def _add_css_class(
      self,
      css_class: str,
      child: Optional[str] = None,
  ) -> str:
    """Adds a CSS class to the control."""
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.classList.add("{css_class}");
        """
    )
    return css_class

  def _remove_css_class(
      self,
      css_class: str,
      child: Optional[str] = None,
  ) -> str:
    """Removes a CSS class from the control."""
    self._run_javascript(
        f"""
        elem = document.getElementById("{self.element_id(child)}");
        elem.classList.remove("{css_class}");
        """
    )
    return css_class


_TLS_TRACKED_SCRIPTS = '__tracked_scripts__'



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representing missing value for a field."""

from typing import Any, Dict
from pyglove.core.object_utils import formatting
from pyglove.core.object_utils import json_conversion


class MissingValue(formatting.Formattable, json_conversion.JSONConvertible):
  """Value placeholder for an unassigned attribute."""

  def format(self, *args, **kwargs):  # pytype: disable=signature-mismatch
    return 'MISSING_VALUE'

  def __ne__(self, other: Any) -> bool:
    return not self.__eq__(other)

  def __eq__(self, other: Any) -> bool:
    return isinstance(other, MissingValue)

  def __hash__(self) -> int:
    return hash(MissingValue.__module__ + MissingValue.__name__)

  def to_json(self, **kwargs) -> Dict[str, Any]:
    return self.to_json_dict(fields=dict(), **kwargs)


# A shortcut global object (constant) for referencing MissingValue.
MISSING_VALUE = MissingValue()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.object_utils.common_traits."""

import unittest
from pyglove.core.object_utils import common_traits


class ExplicitlyOverrideTest(unittest.TestCase):

  def test_explicitly_override(self):
    class A:

      @common_traits.explicit_method_override
      def __init__(self, x, y):
        pass

      def bar(self):
        pass

    common_traits.ensure_explicit_method_override(A.__init__)
    with self.assertRaisesRegex(TypeError, '.* is a PyGlove managed method'):
      common_traits.ensure_explicit_method_override(A.bar)


if __name__ == '__main__':
  unittest.main()



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tab control."""

from typing import Annotated, List, Literal, Union

from pyglove.core.symbolic import object as pg_object
# pylint: disable=g-importing-member
from pyglove.core.views.html.base import Html
from pyglove.core.views.html.base import HtmlConvertible
from pyglove.core.views.html.controls.base import HtmlControl
from pyglove.core.views.html.controls.label import Label
# pylint: enable=g-importing-member


class Tab(pg_object.Object):
  """A tab."""

  label: Annotated[
      Label,
      'The label of the tab.'
  ]

  content: Annotated[
      Union[Html, HtmlConvertible],
      'The content of the tab.'
  ]


@pg_object.use_init_args(
    ['tabs', 'selected', 'tab_position', 'id', 'css_classes', 'styles']
)
class TabControl(HtmlControl):
  """A tab control."""

  tabs: Annotated[
      List[Tab],
      'The tabs of the tab control.'
  ]

  selected: Annotated[
      int,
      'The index of the selected tab.'
  ] = 0

  tab_position: Annotated[
      Literal['top', 'left'],
      'The direction of the tab control.'
  ] = 'top'

  interactive = True

  def _to_html(self, **kwargs):
    return Html.element(
        'div',
        [
            Html.element(
                'div',
                [
                    Html.element(
                        'button',
                        [
                            tab.label
                        ],
                        css_classes=[
                            'tab-button',
                            'selected' if i == self.selected else None
                        ],
                        onclick=(
                            f"""openTab(event, '{self.element_id()}', '{self.element_id(str(i))}')"""
                        )
                    ) for i, tab in enumerate(self.tabs)
                ],
                css_classes=['tab-button-group'],
            ),
        ] + [
            Html.element(
                'div',
                [
                    tab.content
                ],
                css_classes=['tab-content'],
                styles=dict(
                    display='block' if i == self.selected else 'none'
                ),
                id=self.element_id(str(i))
            ) for i, tab in enumerate(self.tabs)
        ],
        css_classes=['tab-control', self.tab_position] + self.css_classes,
        id=self.element_id(),
        styles=self.styles,
    ).add_script(
        """
        function openTab(event, controlId, tabId) {
          const tabButtons = document.querySelectorAll('#' + controlId + '> .tab-button-group > .tab-button');
          for (let i = 0; i < tabButtons.length; i++) {
            tabButtons[i].classList.remove('selected');
          }
          const tabContents = document.querySelectorAll('#' + controlId + '> .tab-content');
          for (let i = 0; i < tabContents.length; i++) {
            tabContents[i].style.display = 'none';
          }
          const tabButton = event.currentTarget;
          tabButton.classList.add('selected');
          document.getElementById(tabId).style.display = "block";
        }
        """
    ).add_style(
        """
        .top .tab-button-group {
          overflow-x: hidden;
          border-bottom: 1px solid #EEE;
        }
        .left .tab-button-group {
          float: left;
          top: 0;
        }
        .tab-button {
          background-color: #EEE;
          border: 1px solid #EEE;
          outline: none;
          cursor: pointer;
          transition: 0.3s;
          padding: 10px 15px 10px 15px;
        }
        .top .tab-button {
          border-top-width: 2px;
        }
        .left .tab-button {
          display: block;
          width: 100%;
          border-left-width: 2px;
        }
        .top .tab-button.selected {
          border-bottom-color: #fff;
          border-top-color: #B721FF;
          background: #fff;
        }
        .left .tab-button.selected {
          border-right-color: #fff;
          border-left-color: #B721FF;
          background: #fff;
        }
        .top .tab-button:hover {
          border-top-color: orange;
        }
        .left .tab-button:hover {
          border-left-color: orange;
        }
        .tab-content {
          display: none;
          padding: 10px;
        }
        .left .tab-content {
          float: left;
        }
        """
    )



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.diff."""

import inspect
from typing import Any
import unittest

from pyglove.core import typing as pg_typing
from pyglove.core.symbolic.dict import Dict
from pyglove.core.symbolic.diff import Diff
from pyglove.core.symbolic.diff import diff as pg_diff
from pyglove.core.symbolic.list import List
from pyglove.core.symbolic.object import members as pg_members
from pyglove.core.symbolic.object import Object


class DiffTest(unittest.TestCase):
  """Tests for `pg.diff`."""

  def test_diff_class(self):

    @pg_members([('x', pg_typing.Any())])
    class A(Object):
      pass

    @pg_members([('x', pg_typing.Any()), ('y', pg_typing.Any())])
    class B(Object):
      pass

    self.assertTrue(bool(Diff(1, 2)))
    self.assertTrue(bool(Diff(A, B)))
    self.assertTrue(bool(Diff(A(1), A(2))))
    self.assertTrue(bool(Diff(A, A, children={'x': Diff(1, 2)})))
    self.assertFalse(bool(Diff()))
    self.assertFalse(bool(Diff(1, 1)))
    self.assertFalse(bool(Diff(A, A)))
    self.assertFalse(bool(Diff(A(1), A(1))))

    self.assertTrue(Diff(1, 2).is_leaf)
    self.assertFalse(
        Diff(int, int, children={'x': Diff(1, 2)}).is_leaf)

    self.assertEqual(Diff(1, 1), 1)
    self.assertEqual(Diff(1, 1), Diff(1, 1))
    self.assertNotEqual(Diff(1, 1), 2)
    self.assertNotEqual(Diff(1, 1), Diff(1, 2))

    self.assertEqual(
        repr(Diff()), 'No diff')
    self.assertEqual(
        repr(Diff(A(1), A(1))), 'A(x=1)')
    self.assertEqual(
        repr(Dict(a=Diff(A(1), A(1)))), '{a=A(x=1)}')
    self.assertEqual(
        repr(Diff(A(1), A(2))), 'Diff(left=A(x=1), right=A(x=2))')
    self.assertEqual(
        repr(Diff(A, A, children={'x': Diff(1, 2)})),
        'A(x=Diff(left=1, right=2))')
    self.assertEqual(
        repr(Diff(A, B, children={
            'x': Diff(1, 2),
            'y': Diff(Diff.MISSING, 3)
        })),
        'A|B(x=Diff(left=1, right=2), y=Diff(left=MISSING, right=3))')
    self.assertEqual(
        repr(Diff(List, List, children={
            '0': Diff(1, 2),
            '1': Diff(Diff.MISSING, 3)
        })),
        '[0=Diff(left=1, right=2), 1=Diff(left=MISSING, right=3)]')

    with self.assertRaisesRegex(
        ValueError,
        '\'left\' must be a type when \'children\' is specified.'):
      Diff(1, int, children={'x': Diff(3, 4)})

    with self.assertRaisesRegex(
        ValueError,
        '\'right\' must be a type when \'children\' is specified.'):
      Diff(int, 2, children={'x': Diff(3, 4)})

    with self.assertRaisesRegex(
        ValueError,
        '\'value\' cannot be accessed when \'left\' and \'right\' '
        'are not the same.'):
      _ = Diff(1, 2).value

  def test_diff_on_simple_types(self):
    self.assertEqual(pg_diff(1, 1), Diff())
    self.assertEqual(pg_diff(1, 1, mode='same'), Diff(1, 1))
    self.assertEqual(pg_diff(1, 1, flatten=True, mode='same'), Diff(1, 1))
    self.assertEqual(pg_diff(1, 2), Diff(1, 2))
    self.assertEqual(pg_diff(1, 2, mode='same'), Diff(1, 2))
    self.assertEqual(pg_diff(1, 2, flatten=True, mode='same'), Diff(1, 2))

  def test_diff_on_list(self):

    @pg_members([('x', pg_typing.Any())])
    class A(Object):
      pass

    # List vs. list.
    self.assertEqual(pg_diff([A(1)], [A(1)]), Diff())
    self.assertEqual(pg_diff([A(1)], [A(1)], mode='same'), Diff([A(1)], [A(1)]))
    self.assertEqual(
        pg_diff([A(1)], [A(0)]),
        Diff(left=List, right=List, children={
            '0': Diff(A, A, children={
                'x': Diff(1, 0)
            })}))

    # List vs. other types.
    self.assertEqual(pg_diff([A(1)], 1), Diff([A(1)], 1))

  def test_diff_on_dict(self):

    @pg_members([('x', pg_typing.Any())])
    class A(Object):
      pass

    # Dict vs. dict.
    self.assertEqual(
        pg_diff({'a': A(1)}, {'a': A(1)}),
        Diff())
    self.assertEqual(
        pg_diff({'a': A(1)}, {'a': A(1)}, mode='same'),
        Diff({'a': A(1)}, {'a': A(1)}))
    self.assertEqual(
        pg_diff({'a': A(1), 'b': A(2), 'c': A(3)},
                {'a': A(1), 'b': A(3), 'd': A(4)}),
        Diff(dict, dict, children={
            'b': Diff(A, A, children={
                'x': Diff(2, 3)
            }),
            'c': Diff(A(3)),
            'd': Diff(right=A(4)),
        }))

    # Dict vs. symbolic object.
    self.assertEqual(
        pg_diff(A(1), {'x': 1}),
        Diff(A(1), {'x': 1}))

    self.assertEqual(
        pg_diff(A(1), {'x': 1}, collapse=True),
        Diff(A, dict))

    self.assertEqual(
        pg_diff(A(1), {'x': 2}, collapse=True),
        Diff(A, dict, children={
            'x': Diff(1, 2)
        }))

    # Dict vs. other types.
    self.assertEqual(pg_diff({'x': 1}, 1), Diff({'x': 1}, 1))

  def test_diff_on_symbolic_object(self):

    @pg_members([('x', pg_typing.Any())])
    class A(Object):
      pass

    @pg_members([('x', pg_typing.Any()), ('y', pg_typing.Any())])
    class B(Object):
      pass

    class C(B):
      pass

    # Same types.
    self.assertEqual(pg_diff(A(1), A(1)), Diff())
    self.assertEqual(pg_diff(A(1), A(1), mode='same'), Diff(A(1), A(1)))
    self.assertEqual(
        pg_diff(B(1, 2), B(1, 3)),
        Diff(B, B, children={'y': Diff(2, 3)}))

    # Different types without collapse.
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 2)),
        Diff(B(1, 2), C(1, 2)))

    # Different types with always collapse.
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 2), collapse=True),
        Diff(B, C))
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 3)),
        pg_diff(B(1, 2), C(1, 3)))
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 3), collapse=True),
        Diff(B, C, children={
            'y': Diff(2, 3)
        }))

  def test_diff_between_symbolic_and_dict(self):

    @pg_members([('x', pg_typing.Any())])
    class A(Object):
      pass

    a = A(A(1))
    d = Dict(x=Dict(x=1))
    self.assertEqual(pg_diff(a, d), Diff(left=a, right=d))
    self.assertEqual(
        pg_diff(a, d, collapse=True),
        Diff(A, Dict, children=dict(x=Diff(A, Dict))))

  def test_collapse(self):

    @pg_members([('x', pg_typing.Any()), ('y', pg_typing.Any())])
    class B(Object):
      pass

    class C(B):
      pass

    @pg_members([('x', pg_typing.Any()), ('y', pg_typing.Any())])
    class D(Object):
      pass

    def collapse_subclass(x, y):
      return issubclass(type(x), type(y)) or issubclass(type(y), type(x))

    self.assertEqual(
        pg_diff(B(1, 2), C(1, 2), collapse=collapse_subclass),
        Diff(B, C))
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 3), collapse=collapse_subclass),
        Diff(B, C, children={
            'y': Diff(2, 3)
        }))
    self.assertEqual(
        pg_diff(B(1, 2), D(1, 3), collapse=collapse_subclass),
        Diff(B(1, 2), D(1, 3)))

    # Test bad collapse option.
    with self.assertRaisesRegex(
        ValueError, 'Unsupported `collapse` value'):
      pg_diff(B(1, 2), C(1, 2), collapse='unsupported_option')

  def test_mode(self):

    @pg_members([('x', pg_typing.Any())])
    class A(Object):
      pass

    @pg_members([('x', pg_typing.Any()), ('y', pg_typing.Any())])
    class B(Object):
      pass

    class C(B):
      pass

    self.assertEqual(pg_diff(A(1), A(1), mode='diff'), Diff())
    self.assertEqual(pg_diff(A(1), A(1), mode='same'), Diff(A(1), A(1)))
    self.assertEqual(pg_diff(A(1), A(1), mode='both'), Diff(A(1), A(1)))

    self.assertEqual(
        pg_diff(B(1, 2), B(1, 3), mode='diff'),
        Diff(B, B, children={'y': Diff(2, 3)}))
    self.assertEqual(
        pg_diff(B(1, 2), B(1, 3), mode='same'),
        Diff(B, B, children={'x': Diff(1, 1)}))
    self.assertEqual(
        pg_diff(B(1, 2), B(2, 2), mode='both'),
        Diff(B, B, children={
            'x': Diff(1, 2),
            'y': Diff(2, 2)
        }))

    self.assertEqual(
        pg_diff(B(1, 2), C(1, 2), collapse=True, mode='diff'), Diff(B, C))
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 2), collapse=True, mode='same'),
        Diff(B, C, children={
            'x': Diff(1, 1),
            'y': Diff(2, 2)
        }))
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 2), collapse=True, mode='both'),
        Diff(B, C, children={
            'x': Diff(1, 1),
            'y': Diff(2, 2)
        }))

    self.assertEqual(
        pg_diff(B(1, 2), C(1, 3), collapse=True, mode='diff'),
        Diff(B, C, children={
            'y': Diff(2, 3)
        }))
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 3), collapse=True, mode='same'),
        Diff(B, C, children={
            'x': Diff(1, 1)
        }))
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 3), collapse=True, mode='both'),
        Diff(B, C, children={
            'x': Diff(1, 1),
            'y': Diff(2, 3)
        }))

  def test_flatten(self):

    @pg_members([('x', pg_typing.Any())])
    class A(Object):
      pass

    @pg_members([('x', pg_typing.Any()), ('y', pg_typing.Any())])
    class B(Object):
      pass

    class C(B):
      pass

    self.assertEqual(
        pg_diff(A(Dict(a=1, b=2, c=3)), A(Dict(a=1, b=3, d=4)), flatten=True),
        {
            'x.b': Diff(2, 3),
            'x.c': Diff(3, Diff.MISSING),
            'x.d': Diff(Diff.MISSING, 4)
        })
    self.assertEqual(
        pg_diff(B(1, 2), C(1, 3), collapse=True, flatten=True),
        {
            'y': Diff(2, 3),
            '_type': Diff(B, C),
        })

  def test_to_html(self):

    class Foo(Object):
      x: Any
      y: Any

    class Bar(Foo):
      pass

    def assert_style(html, expected):
      expected = inspect.cleandoc(expected).strip()
      actual = html.style_section.strip()
      if expected not in actual:
        print(actual)
      self.assertIn(expected, actual)

    def assert_content(html, expected):
      expected = inspect.cleandoc(expected).strip()
      actual = html.content.strip()
      if actual != expected:
        print(actual)
      self.assertEqual(actual.strip(), expected)

    assert_style(
        pg_diff(1, 1).to_html(),
        """
        /* Diff styles. */
        .has-diff.summary-title::after {
          content: ' (diff)';
          color: #aaa;
        }
        .has-diff.summary-title {
          background-color: yellow;
        }
        .no-diff.summary-title::after {
          content: ' (no diff)';
          font-style: italic;
          font-weight: normal;
          color: #aaa;
        }
        .no-diff {
          opacity: 0.6;
        }
        .no-diff.simple_value::after {
          content: '(no diff)';
          margin-left: 0.5em;
          color: #aaa;
          font-style: italic;
        }
        .diff-empty::before {
          content: '(empty)';
          font-style: italic;
          margin-left: 0.5em;
          color: #aaa;
        }
        .diff-left.summary-title::before, .diff-left.simple-value::before{
          content: '🇱';
        }
        .diff-left {
          background-color: #ffcccc;
        }
        .diff-right.summary-title::before, .diff-right.simple-value::before{
          content: '🇷';
        }
        .diff-right {
          background-color: #ccffcc;
        }
        """
    )

    # No diff leaf value (diff only)
    assert_content(
        pg_diff(1, 1).to_html(),
        """
        <details open class="pyglove diff no-diff"><summary><div class="summary-title no-diff">Diff</div><span class="tooltip no-diff">No diff</span></summary><span class="diff-empty"></span></details>
        """
    )

    # No diff leaf value (keep both)
    assert_content(
        pg_diff(1, 1, mode='both').to_html(),
        """
        <span class="simple-value int no-diff">1</span>
        """
    )
    # No diff complex value (diff only)
    assert_content(
        pg_diff(
            Foo(x=1, y=[Foo(x=2, y=3)]),
            Foo(x=1, y=[Foo(x=2, y=3)])
        ).to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove diff no-diff"><summary><div class="summary-title no-diff">Diff</div></summary><span class="diff-empty"></span></details>
        """
    )
    # No diff complex value.
    assert_content(
        pg_diff(
            Foo(x=1, y=[Foo(x=2, y=3)]),
            Foo(x=1, y=[Foo(x=2, y=3)]), mode='both'
        ).to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove diff no-diff"><summary><div class="summary-title no-diff">Foo(...)</div></summary><div class="complex-value foo no-diff"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details class="pyglove list"><summary><div class="summary-name">y</div><div class="summary-title">List(...)</div></summary><div class="complex-value list"><table><tr><td><span class="object-key int">0</span></td><td><details class="pyglove foo"><summary><div class="summary-title">Foo(...)</div></summary><div class="complex-value foo"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details><details open class="pyglove int"><summary><div class="summary-name">y</div><div class="summary-title">int</div></summary><span class="simple-value int">3</span></details></div></details></td></tr></table></div></details></div></details>
        """
    )

    # Diff on simple value.
    assert_content(
        pg_diff(1, 2).to_html(),
        """
        <div class="diff-value"><span class="simple-value int diff-left">1</span><span class="simple-value int diff-right">2</span></div>
        """
    )

    # Diff on list.
    assert_content(
        pg_diff([0, 1, 2], [0, 2], mode='both').to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove diff has-diff"><summary><div class="summary-title has-diff">List</div></summary><div class="complex-value list-class"><table><tr><td><span class="object-key int no-diff">0</span><span class="tooltip">[0]</span></td><td><span class="simple-value int no-diff">0</span></td></tr><tr><td><span class="object-key int">1</span><span class="tooltip">[1]</span></td><td><div class="diff-value"><span class="simple-value int diff-left">1</span><span class="simple-value int diff-right">2</span></div></td></tr><tr><td><span class="object-key int">2</span><span class="tooltip">[2]</span></td><td><div class="diff-value"><span class="simple-value int diff-left">2</span></div></td></tr></table></div></details>
        """
    )

    # Diff on dict.
    assert_content(
        pg_diff(dict(x=1, y=2, z=3), dict(x=1, y=3, w=4), mode='both').to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove diff has-diff"><summary><div class="summary-title has-diff">dict</div></summary><div class="complex-value dict-class"><table><tr><td><span class="object-key str no-diff">x</span><span class="tooltip">x</span></td><td><span class="simple-value int no-diff">1</span></td></tr><tr><td><span class="object-key str">y</span><span class="tooltip">y</span></td><td><div class="diff-value"><span class="simple-value int diff-left">2</span><span class="simple-value int diff-right">3</span></div></td></tr><tr><td><span class="object-key str">z</span><span class="tooltip">z</span></td><td><div class="diff-value"><span class="simple-value int diff-left">3</span></div></td></tr><tr><td><span class="object-key str">w</span><span class="tooltip">w</span></td><td><div class="diff-value"><span class="simple-value int diff-right">4</span></div></td></tr></table></div></details>
        """
    )

    # Diff on symbolic objects of the same type.
    assert_content(
        pg_diff(
            Foo(x=2, y=Foo(x=3, y=3)),
            Foo(x=1, y=Foo(x=2, y=3)),
            mode='both',
        ).to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove diff has-diff"><summary><div class="summary-title has-diff">Foo</div></summary><div class="complex-value foo-class"><table><tr><td><span class="object-key str">x</span><span class="tooltip">x</span></td><td><div class="diff-value"><span class="simple-value int diff-left">2</span><span class="simple-value int diff-right">1</span></div></td></tr><tr><td><span class="object-key str">y</span><span class="tooltip">y</span></td><td><details open class="pyglove diff has-diff"><summary><div class="summary-title has-diff">Foo</div></summary><div class="complex-value foo-class"><table><tr><td><span class="object-key str">x</span><span class="tooltip">y.x</span></td><td><div class="diff-value"><span class="simple-value int diff-left">3</span><span class="simple-value int diff-right">2</span></div></td></tr><tr><td><span class="object-key str no-diff">y</span><span class="tooltip">y.y</span></td><td><span class="simple-value int no-diff">3</span></td></tr></table></div></details></td></tr></table></div></details>
        """
    )

    # Diff on symbolic objects of different types.
    assert_content(
        pg_diff(
            Foo(x=2, y=Foo(x=3, y=3)),
            Bar(x=2, y=Foo(x=3, y=3)),
            mode='both',
        ).to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <div class="diff-value"><details open class="pyglove foo diff-left"><summary><div class="summary-title diff-left">Foo(...)</div></summary><div class="complex-value foo"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details><details class="pyglove foo"><summary><div class="summary-name">y</div><div class="summary-title">Foo(...)</div></summary><div class="complex-value foo"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">3</span></details><details open class="pyglove int"><summary><div class="summary-name">y</div><div class="summary-title">int</div></summary><span class="simple-value int">3</span></details></div></details></div></details><details open class="pyglove bar diff-right"><summary><div class="summary-title diff-right">Bar(...)</div></summary><div class="complex-value bar"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details><details class="pyglove foo"><summary><div class="summary-name">y</div><div class="summary-title">Foo(...)</div></summary><div class="complex-value foo"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">3</span></details><details open class="pyglove int"><summary><div class="summary-name">y</div><div class="summary-title">int</div></summary><span class="simple-value int">3</span></details></div></details></div></details></div>
        """
    )

    # Different types but same values, with value collapsing.
    assert_content(
        pg_diff(
            Foo(x=2, y=Foo(x=3, y=3)),
            Bar(x=2, y=Bar(x=3, y=3)),
            mode='both', collapse=True,
        ).to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove diff has-diff"><summary><div class="summary-title has-diff">Foo | Bar</div></summary><div class="complex-value foo-class"><table><tr><td><span class="object-key str no-diff">x</span><span class="tooltip">x</span></td><td><span class="simple-value int no-diff">2</span></td></tr><tr><td><span class="object-key str">y</span><span class="tooltip">y</span></td><td><details open class="pyglove diff has-diff"><summary><div class="summary-title has-diff">Foo | Bar</div></summary><div class="complex-value foo-class"><table><tr><td><span class="object-key str no-diff">x</span><span class="tooltip">y.x</span></td><td><span class="simple-value int no-diff">3</span></td></tr><tr><td><span class="object-key str no-diff">y</span><span class="tooltip">y.y</span></td><td><span class="simple-value int no-diff">3</span></td></tr></table></div></details></td></tr></table></div></details>
        """
    )

    # Different types and different values, with value collapsing.
    assert_content(
        pg_diff(
            Foo(x=2, y=Foo(x=3, y=3)),
            Bar(x=3, y=Bar(x=2, y=3)),
            mode='both', collapse=True,
        ).to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
        ),
        """
        <details open class="pyglove diff has-diff"><summary><div class="summary-title has-diff">Foo | Bar</div></summary><div class="complex-value foo-class"><table><tr><td><span class="object-key str">x</span><span class="tooltip">x</span></td><td><div class="diff-value"><span class="simple-value int diff-left">2</span><span class="simple-value int diff-right">3</span></div></td></tr><tr><td><span class="object-key str">y</span><span class="tooltip">y</span></td><td><details open class="pyglove diff has-diff"><summary><div class="summary-title has-diff">Foo | Bar</div></summary><div class="complex-value foo-class"><table><tr><td><span class="object-key str">x</span><span class="tooltip">y.x</span></td><td><div class="diff-value"><span class="simple-value int diff-left">3</span><span class="simple-value int diff-right">2</span></div></td></tr><tr><td><span class="object-key str no-diff">y</span><span class="tooltip">y.y</span></td><td><span class="simple-value int no-diff">3</span></td></tr></table></div></details></td></tr></table></div></details>
        """
    )

    # Diff with uncollapsing UI.
    assert_content(
        pg_diff(
            [
                Foo(1, 2), Foo(1, 2), None,
                Foo(x=1, y=Foo(2, 3)), [dict(x=1, y=2)]
            ],
            [
                Foo(1, 2), Bar(1, 2), dict(x=1),
                [1, 2], Foo(x=2, y=Foo(2, 4)), [dict(x=3)]
            ],
            mode='both'
        ).to_html(
            enable_summary_tooltip=False,
            enable_key_tooltip=False,
            uncollapse=['[0]', '[1].right', '[3].left.y'],
        ),
        """
        <details open class="pyglove diff has-diff"><summary><div class="summary-title has-diff">List</div></summary><div class="complex-value list-class"><table><tr><td><span class="object-key int no-diff">0</span><span class="tooltip">[0]</span></td><td><details open class="pyglove diff no-diff"><summary><div class="summary-title no-diff">Foo(...)</div></summary><div class="complex-value foo no-diff"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove int"><summary><div class="summary-name">y</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details></div></details></td></tr><tr><td><span class="object-key int">1</span><span class="tooltip">[1]</span></td><td><div class="diff-value"><details open class="pyglove foo diff-left"><summary><div class="summary-title diff-left">Foo(...)</div></summary><div class="complex-value foo"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove int"><summary><div class="summary-name">y</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details></div></details><details open class="pyglove bar diff-right"><summary><div class="summary-title diff-right">Bar(...)</div></summary><div class="complex-value bar"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove int"><summary><div class="summary-name">y</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details></div></details></div></td></tr><tr><td><span class="object-key int">2</span><span class="tooltip">[2]</span></td><td><div class="diff-value"><span class="simple-value none-type diff-left">None</span><details open class="pyglove dict diff-right"><summary><div class="summary-title diff-right">Dict(...)</div></summary><div class="complex-value dict"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details></div></details></div></td></tr><tr><td><span class="object-key int">3</span><span class="tooltip">[3]</span></td><td><div class="diff-value"><details open class="pyglove foo diff-left"><summary><div class="summary-title diff-left">Foo(...)</div></summary><div class="complex-value foo"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove foo"><summary><div class="summary-name">y</div><div class="summary-title">Foo(...)</div></summary><div class="complex-value foo"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details><details open class="pyglove int"><summary><div class="summary-name">y</div><div class="summary-title">int</div></summary><span class="simple-value int">3</span></details></div></details></div></details><details open class="pyglove list diff-right"><summary><div class="summary-title diff-right">List(...)</div></summary><div class="complex-value list"><table><tr><td><span class="object-key int">0</span></td><td><span class="simple-value int">1</span></td></tr><tr><td><span class="object-key int">1</span></td><td><span class="simple-value int">2</span></td></tr></table></div></details></div></td></tr><tr><td><span class="object-key int">4</span><span class="tooltip">[4]</span></td><td><div class="diff-value"><details open class="pyglove list diff-left"><summary><div class="summary-title diff-left">List(...)</div></summary><div class="complex-value list"><table><tr><td><span class="object-key int">0</span></td><td><details class="pyglove dict"><summary><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">1</span></details><details open class="pyglove int"><summary><div class="summary-name">y</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details></div></details></td></tr></table></div></details><details open class="pyglove foo diff-right"><summary><div class="summary-title diff-right">Foo(...)</div></summary><div class="complex-value foo"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details><details class="pyglove foo"><summary><div class="summary-name">y</div><div class="summary-title">Foo(...)</div></summary><div class="complex-value foo"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">2</span></details><details open class="pyglove int"><summary><div class="summary-name">y</div><div class="summary-title">int</div></summary><span class="simple-value int">4</span></details></div></details></div></details></div></td></tr><tr><td><span class="object-key int">5</span><span class="tooltip">[5]</span></td><td><div class="diff-value"><details open class="pyglove list diff-right"><summary><div class="summary-title diff-right">List(...)</div></summary><div class="complex-value list"><table><tr><td><span class="object-key int">0</span></td><td><details class="pyglove dict"><summary><div class="summary-title">Dict(...)</div></summary><div class="complex-value dict"><details open class="pyglove int"><summary><div class="summary-name">x</div><div class="summary-title">int</div></summary><span class="simple-value int">3</span></details></div></details></td></tr></table></div></details></div></td></tr></table></div></details>
        """
    )


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Symbolic dict."""

import typing
from typing import Any, Callable, Iterable, Iterator, List, Optional, Sequence, Set, Tuple, Union

from pyglove.core import object_utils
from pyglove.core import typing as pg_typing
from pyglove.core.symbolic import base
from pyglove.core.symbolic import flags


class Dict(dict, base.Symbolic, pg_typing.CustomTyping):
  """Symbolic dict.

  ``pg.Dict`` implements a dict type whose instances are symbolically
  programmable, which is a subclass of the built-in Python ``dict`` and
  a subclass of :class:`pyglove.Symbolic`.

  ``pg.Dict`` provides the following features:

   * It a symbolic programmable dict with string keys.
   * It enables attribute access on dict keys.
   * It supports symbolic validation and value completitions based on schema.
   * It provides events to handle sub-nodes changes.

  ``pg.Dict`` can be used as a regular dict with string keys::

    # Construct a symbolic dict from key value pairs.
    d = pg.Dict(x=1, y=2)

  or::

    # Construct a symbolic dict from a mapping object.
    d = pg.Dict({'x': 1, 'y': 2})

  Besides regular items access using ``[]``, it allows attribute access
  to its keys::

    # Read access to key `x`.
    assert d.x == 1

    # Write access to key 'y'.
    d.y = 1

  ``pg.Dict`` supports symbolic validation when the ``value_spec`` argument
  is provided::

    d = pg.Dict(x=1, y=2, value_spec=pg.typing.Dict([
        ('x', pg.typing.Int(min_value=1)),
        ('y', pg.typing.Int(min_value=1)),
        (pg.typing.StrKey('foo.*'), pg.typing.Str())
    ])

    # Okay: all keys started with 'foo' is acceptable and are strings.
    d.foo1 = 'abc'

    # Raises: 'bar' is not acceptable as keys in the dict.
    d.bar = 'abc'

  Users can mutate the values contained in it::

    d = pg.Dict(x=pg.Dict(y=1), p=pg.List([0]))
    d.rebind({
      'x.y': 2,
      'p[0]': 1
    })

  It also allows the users to subscribe subtree updates::

    def on_change(updates):
      print(updates)

    d = pg.Dict(x=1, onchange_callaback=on_change)

    # `on_change` will be triggered on item insertion.
    d['y'] = {'z': 1}

    # `on_change` will be triggered on item removal.
    del d.x

    # `on_change` will also be triggered on subtree change.
    d.rebind({'y.z': 2})
  """

  @classmethod
  def partial(cls,
              dict_obj: Optional[typing.Dict[str, Any]] = None,
              value_spec: Optional[pg_typing.Dict] = None,
              *,
              onchange_callback: Optional[Callable[
                  [typing.Dict[object_utils.KeyPath, base.FieldUpdate]], None]
              ] = None,  # pylint: disable=bad-continuation
              **kwargs) -> 'Dict':
    """Class method that creates a partial Dict object."""
    return cls(dict_obj,
               value_spec=value_spec,
               onchange_callback=onchange_callback,
               allow_partial=True,
               **kwargs)

  @classmethod
  def from_json(cls,
                json_value: Any,
                *,
                value_spec: Optional[pg_typing.Dict] = None,
                allow_partial: bool = False,
                root_path: Optional[object_utils.KeyPath] = None,
                **kwargs) -> 'Dict':
    """Class method that load an symbolic Dict from a JSON value.

    Args:
      json_value: Input JSON value, only JSON dict is acceptable.
      value_spec: An optional value spec to apply.
      allow_partial: Whether to allow members of the dict to be partial.
      root_path: KeyPath of loaded object in its object tree.
      **kwargs: Allow passing through keyword arguments that are not applicable.

    Returns:
      A schemaless symbolic dict. For example::

        d = Dict.from_json({
          'a': {
            '_type': '__main__.Foo',
            'f1': 1,
            'f2': {
              'f21': True
            }
          }
        })

        assert d.value_spec is None
        # Okay:
        d.b = 1

        # a.f2 is bound by class Foo's field 'f2' definition (assume it defines
        # a schema for the Dict field).
        assert d.a.f2.value_spec is not None

        # Not okay:
        d.a.f2.abc = 1
    """
    return cls(
        {
            k: base.from_json(
                v,
                root_path=object_utils.KeyPath(k, root_path),
                allow_partial=allow_partial,
                **kwargs
            ) for k, v in json_value.items()
        },
        value_spec=value_spec,
        root_path=root_path,
        allow_partial=allow_partial,
    )

  def __init__(self,
               dict_obj: Union[
                   None,
                   Iterable[Tuple[str, Any]],
                   typing.Dict[str, Any]] = None,
               *,
               value_spec: Optional[pg_typing.Dict] = None,
               onchange_callback: Optional[Callable[
                   [typing.Dict[object_utils.KeyPath, base.FieldUpdate]], None]
               ] = None,  # pylint: disable=bad-continuation
               **kwargs):
    """Constructor.

    Args:
      dict_obj: A dict as initial value for this Dict.
      value_spec: Value spec that applies to this Dict.
      onchange_callback: Callback when sub-tree has been modified.
      **kwargs: Key value pairs that will be inserted into the dict as initial
        value, which provides a syntax sugar for usage as below: d =
          pg.Dict(a=1, b=2)
    """
    if value_spec and not isinstance(value_spec, pg_typing.Dict):
      raise TypeError(
          f'Argument \'value_spec\' must be a `pg.typing.Dict` object. '
          f'Encountered {value_spec}')

    allow_partial = kwargs.pop('allow_partial', False)
    accessor_writable = kwargs.pop('accessor_writable', True)
    sealed = kwargs.pop('sealed', False)
    root_path = kwargs.pop('root_path', None)

    # Skip schema check when dict_obj is validated against
    # schema externally. This flag is helpful to avoid duplicated schema
    # check in nested structures, which takes effect only when value_spec
    # is not None.
    pass_through = kwargs.pop('pass_through', False)

    # If True, the parent of dict items should be set to `self.sym_parent`,
    # This is useful when Dict is used as the field container of
    # pg.Object.
    self._set_raw_attr(
        '_as_object_attributes_container',
        kwargs.pop('as_object_attributes_container', False),
    )

    # We copy the symbolic form of dict values instead of their evaluated
    # values.
    if isinstance(dict_obj, Dict):
      dict_obj = {k: v for k, v in dict_obj.sym_items()}
    elif dict_obj is not None:
      dict_obj = dict(dict_obj)

    # NOTE(daiyip): we call __init__ of superclasses explicitly instead of
    # calling super().__init__(...) since dict.__init__ does
    # not follow super(...).__init__ fashion, which will lead to
    # Symbolic.__init__ uncalled.
    base.Symbolic.__init__(
        self,
        allow_partial=allow_partial,
        accessor_writable=True,
        # We delay seal operation until members are filled.
        sealed=False,
        root_path=root_path)

    dict.__init__(self)
    self._value_spec = None
    self._onchange_callback = None

    # NOTE(daiyip): values in kwargs is prior to dict_obj.
    dict_obj = dict_obj or {}
    for k, v in kwargs.items():
      dict_obj[k] = v

    if value_spec:
      if pass_through:
        for k, v in dict_obj.items():
          super().__setitem__(k, self._relocate_if_symbolic(k, v))

        # NOTE(daiyip): when pass_through is on, we simply trust input
        # dict is validated and filled with values of their final form (
        # symbolic Dict/List vs. dict/list). This prevents members from
        # repeated validation and transformation.
        self._value_spec = value_spec
      else:
        for k, v in dict_obj.items():
          super().__setitem__(k, self._formalized_value(k, None, v))
        self.use_value_spec(value_spec, allow_partial)
    else:
      for k, v in dict_obj.items():
        self._set_item_without_permission_check(k, v)

    # NOTE(daiyip): We set onchange callback at the end of init to avoid
    # triggering during initialization.
    self._onchange_callback = onchange_callback
    self.set_accessor_writable(accessor_writable)
    self.seal(sealed)

  @property
  def value_spec(self) -> Optional[pg_typing.Dict]:
    """Returns value spec of this dict.

    NOTE(daiyip): If this dict is schema-less, value_spec will be None.
    """
    return self._value_spec

  def use_value_spec(self,
                     value_spec: Optional[pg_typing.Dict],
                     allow_partial: bool = False) -> 'Dict':
    """Applies a ``pg.typing.Dict`` as the value spec for current dict.

    Args:
      value_spec: A Dict ValueSpec to apply to this Dict.
        If current Dict is schema-less (whose immediate members are not
        validated against schema), and `value_spec` is not None, the value spec
        will be applied to the Dict.
        Or else if current Dict is already symbolic (whose immediate members
        are under the constraint of a Dict value spec), and `value_spec` is
        None, current Dict will become schema-less. However, the schema
        constraints for non-immediate members will remain.
      allow_partial: Whether allow partial dict based on the schema. This flag
        will override allow_partial flag in __init__ for spec-less Dict.

    Returns:
      Self.

    Raises:
      ValueError: validation failed due to value error.
      RuntimeError: Dict is already bound with another spec.
      TypeError: type errors during validation.
      KeyError: key errors during validation.
    """
    if value_spec is None:
      self._value_spec = None
      self._accessor_writable = True
      return self

    if not isinstance(value_spec, pg_typing.Dict):
      raise ValueError(
          self._error_message(
              f'Value spec for list must be a `pg.typing.Dict` object. '
              f'Encountered: {value_spec!r}'))

    if self._value_spec and self._value_spec != value_spec:
      raise RuntimeError(
          self._error_message(
              f'Dict is already bound with a different value spec: '
              f'{self._value_spec}. New value spec: {value_spec}.'))

    self._allow_partial = allow_partial

    if flags.is_type_check_enabled():
      # NOTE(daiyip): self._value_spec will be set in Dict.custom_apply method
      # called by value_spec.apply, thus we don't need to set self._value_spec
      # explicitly.
      value_spec.apply(
          self,
          allow_partial=base.accepts_partial(self),
          child_transform=base.symbolic_transform_fn(self._allow_partial),
          root_path=self.sym_path)
    else:
      self._value_spec = value_spec
    return self

  def _sym_parent_for_children(self) -> Optional[base.Symbolic]:
    if self._as_object_attributes_container:
      return self.sym_parent
    return self

  def _sym_rebind(
      self, path_value_pairs: typing.Dict[object_utils.KeyPath, Any]
      ) -> List[base.FieldUpdate]:
    """Subclass specific rebind implementation."""
    updates = []
    for k, v in path_value_pairs.items():
      update = self._set_item_of_current_tree(k, v)
      if update is not None:
        updates.append(update)
    return updates

  def _sym_missing(self) -> typing.Dict[str, Any]:
    """Returns missing values.

    Returns:
      A dict of key to MISSING_VALUE.
    """
    missing = dict()
    if self._value_spec and self._value_spec.schema:
      matched_keys, _ = self._value_spec.schema.resolve(self.keys())
      for key_spec, keys in matched_keys.items():
        field = self._value_spec.schema[key_spec]
        assert keys or isinstance(key_spec, pg_typing.NonConstKey), key_spec
        if keys:
          for key in keys:
            v = self.sym_getattr(key)
            if object_utils.MISSING_VALUE == v:
              missing[key] = field.value.default
            else:
              if isinstance(v, base.Symbolic):
                missing_child = v.sym_missing(flatten=False)
                if missing_child:
                  missing[key] = missing_child
    else:
      for k, v in self.sym_items():
        if isinstance(v, base.Symbolic):
          missing_child = v.sym_missing(flatten=False)
          if missing_child:
            missing[k] = missing_child
    return missing

  def _sym_nondefault(self) -> typing.Dict[str, Any]:
    """Returns non-default values as key/value pairs in a dict."""
    non_defaults = dict()
    if self._value_spec is not None and self._value_spec.schema:
      dict_schema = self._value_spec.schema
      matched_keys, _ = dict_schema.resolve(self.keys())
      for key_spec, keys in matched_keys.items():
        value_spec = dict_schema[key_spec].value
        for key in keys:
          diff = self._diff_base(self.sym_getattr(key), value_spec.default)
          if pg_typing.MISSING_VALUE != diff:
            non_defaults[key] = diff
    else:
      for k, v in self.sym_items():
        if isinstance(v, base.Symbolic):
          non_defaults_child = v.sym_nondefault(flatten=False)
          if non_defaults_child:
            non_defaults[k] = non_defaults_child
        else:
          non_defaults[k] = v
    return non_defaults

  def _diff_base(self, value: Any, base_value: Any) -> Any:
    """Computes the diff between a value and a base value."""
    if base.eq(value, base_value):
      return pg_typing.MISSING_VALUE

    if (isinstance(value, list)
        or not isinstance(value, base.Symbolic)
        or pg_typing.MISSING_VALUE == base_value):
      return value

    if value.__class__ is base_value.__class__:
      getter = lambda x, k: x.sym_getattr(k, pg_typing.MISSING_VALUE)
    elif isinstance(value, dict) and isinstance(base_value, dict):
      getter = lambda x, k: x.get(k, pg_typing.MISSING_VALUE)
    else:
      return value

    diff = {}
    for k, v in value.sym_items():
      base_v = getter(base_value, k)
      child_diff = self._diff_base(v, base_v)
      if pg_typing.MISSING_VALUE != child_diff:
        diff[k] = child_diff
    return diff

  def seal(self, sealed: bool = True) -> 'Dict':
    """Seals or unseals current object from further modification."""
    if self.is_sealed == sealed:
      return self
    for v in self.sym_values():
      if isinstance(v, base.Symbolic):
        v.seal(sealed)
    super().seal(sealed)
    return self

  def sym_attr_field(
      self, key: Union[str, int]
      ) -> Optional[pg_typing.Field]:
    """Returns the field definition for a symbolic attribute."""
    if self._value_spec is None or self._value_spec.schema is None:
      return None
    return self._value_spec.schema.get_field(key)  # pytype: disable=attribute-error

  def sym_hasattr(self, key: Union[str, int]) -> bool:
    """Tests if a symbolic attribute exists."""
    return key in self

  def sym_keys(self) -> Iterator[str]:
    """Iterates the keys of symbolic attributes."""
    if self._value_spec is None or self._value_spec.schema is None:
      for key in super().__iter__():
        yield key
    else:
      traversed = set()
      for key_spec in self._value_spec.schema.keys():  # pytype: disable=attribute-error
        if isinstance(key_spec, pg_typing.ConstStrKey) and key_spec in self:
          yield key_spec.text
          traversed.add(key_spec.text)

      if len(traversed) < len(self):
        for key in super().__iter__():
          if key not in traversed:
            yield key

  def sym_values(self) -> Iterator[Any]:
    """Iterates the values of symbolic attributes."""
    for k in self.sym_keys():
      yield self._sym_getattr(k)

  def sym_items(self) -> Iterator[
      Tuple[str, Any]]:
    """Iterates the (key, value) pairs of symbolic attributes."""
    for k in self.sym_keys():
      yield k, self._sym_getattr(k)

  def sym_setparent(self, parent: base.Symbolic):
    """Override set parent of Dict to handle the passing through scenario."""
    super().sym_setparent(parent)
    # NOTE(daiyip): when flag `as_object_attributes_container` is on, it sets
    # the parent of child symbolic values using its parent.
    if self._as_object_attributes_container:
      for v in self.sym_values():
        if isinstance(v, base.TopologyAware):
          v.sym_setparent(parent)

  def sym_hash(self) -> int:
    """Symbolic hashing."""
    return base.sym_hash(
        (self.__class__,
         tuple([(k, base.sym_hash(v)) for k, v in self.sym_items()
                if v != pg_typing.MISSING_VALUE])))

  def _sym_getattr(  # pytype: disable=signature-mismatch  # overriding-parameter-type-checks
      self, key: str) -> Any:
    """Gets symbolic attribute by key."""
    return super().__getitem__(key)

  def _sym_clone(self, deep: bool, memo=None) -> 'Dict':
    """Override Symbolic._sym_clone."""
    source = dict()
    for k, v in self.sym_items():
      if deep or isinstance(v, base.Symbolic):
        v = base.clone(v, deep, memo)
      source[k] = v
    return Dict(
        source,
        value_spec=self._value_spec,
        allow_partial=self._allow_partial,
        accessor_writable=self._accessor_writable,
        sealed=self._sealed,
        onchange_callback=self._onchange_callback,
        # NOTE(daiyip): parent and root_path are reset to empty
        # for copy object.
        root_path=None,
        pass_through=True)

  def _update_children_paths(
      self,
      old_path: object_utils.KeyPath,
      new_path: object_utils.KeyPath) -> None:
    """Update children paths according to root_path of current node."""
    del old_path
    for k, v in self.sym_items():
      if isinstance(v, base.TopologyAware):
        v.sym_setpath(object_utils.KeyPath(k, new_path))

  def _set_item_without_permission_check(  # pytype: disable=signature-mismatch  # overriding-parameter-type-checks
      self, key: str, value: Any) -> Optional[base.FieldUpdate]:
    """Set item without permission check."""
    if not isinstance(key, str):
      raise KeyError(self._error_message(
          f'Key must be string type. Encountered {key!r}.'))

    old_value = self.get(key, pg_typing.MISSING_VALUE)
    if old_value is value:
      return None

    field = None
    if self._value_spec and self._value_spec.schema:
      field = self._value_spec.schema.get_field(key)
      if not field:
        if (self.sym_parent is not None
            and self.sym_parent.sym_path == self.sym_path):
          container_cls = self.sym_parent.__class__
        else:
          container_cls = self.__class__
        raise KeyError(
            self._error_message(
                f'Key \'{key}\' is not allowed for {container_cls}.'))

    # Detach old value from object tree.
    if isinstance(old_value, base.TopologyAware):
      old_value.sym_setparent(None)
      old_value.sym_setpath(object_utils.KeyPath())

    if (pg_typing.MISSING_VALUE == value and
        (not field or isinstance(field.key, pg_typing.NonConstKey))):
      if key in self:
        # Using pg.MISSING_VALUE for deleting keys.
        super().__delitem__(key)
        new_value = pg_typing.MISSING_VALUE
      else:
        # This condition could trigger when copying a partial Dict to a Dict
        # without schema.
        return None
    else:
      new_value = self._formalized_value(key, field, value)
      super().__setitem__(key, new_value)

    # NOTE(daiyip): If current dict is the field dict of a symbolic object,
    # Use parent object as update target.
    target = self
    if (self.sym_parent is not None
        and self.sym_parent.sym_path == self.sym_path):
      target = self.sym_parent
    return base.FieldUpdate(
        self.sym_path + key, target, field, old_value, new_value)

  def _formalized_value(self, name: str,
                        field: Optional[pg_typing.Field],
                        value: Any) -> Any:
    """Get transformed (formal) value from user input."""
    allow_partial = base.accepts_partial(self)
    if field and pg_typing.MISSING_VALUE == value:
      # NOTE(daiyip): default value is already in transformed form.
      value = field.default_value
    else:
      value = base.from_json(
          value,
          allow_partial=allow_partial,
          root_path=object_utils.KeyPath(name, self.sym_path))
    if field and flags.is_type_check_enabled():
      value = field.apply(
          value,
          allow_partial=allow_partial,
          transform_fn=base.symbolic_transform_fn(self._allow_partial),
          root_path=object_utils.KeyPath(name, self.sym_path))
    return self._relocate_if_symbolic(name, value)

  @property
  def _subscribes_field_updates(self) -> bool:
    """Returns True if current dict subscribes field updates."""
    return self._onchange_callback is not None

  def _on_change(self, field_updates: typing.Dict[object_utils.KeyPath,
                                                  base.FieldUpdate]):
    """On change event of Dict."""
    if self._onchange_callback:
      self._onchange_callback(field_updates)

  def _init_kwargs(self) -> typing.Dict[str, Any]:
    kwargs = super()._init_kwargs()
    if not self._accessor_writable:
      kwargs['accessor_writable'] = False
    if self._onchange_callback is not None:
      kwargs['onchange_callback'] = self._onchange_callback
    # NOTE(daiyip): We do not serialize ValueSpec for now as in most use
    # cases they come from the subclasses of `pg.Object`.
    return kwargs

  def __getstate__(self) -> Any:
    """Customizes pickle.dump."""
    return dict(value=dict(self), kwargs=self._init_kwargs())

  def __setstate__(self, state) -> None:
    """Customizes pickle.load."""
    self.__init__(state['value'], **state['kwargs'])

  def __getitem__(self, key: str) -> Any:
    """Get item in this Dict."""
    try:
      return self.sym_inferred(key)
    except AttributeError as e:
      raise KeyError(key) from e

  def __setitem__(self, key: str, value: Any) -> None:
    """Set item in this Dict.

    Args:
      key: String key. (Please be noted that key path is not supported.)
      value: Value to be inserted.

    Raises:
      WritePermissionError: when Dict cannot be modified by accessor or
        is sealed.
      KeyError: Key is not allowed according to the value spec.
      ValueError: Value is not acceptable according to the value spec.
    """
    # NOTE(daiyip): THIS IS A WORKAROUND FOR WORKING WITH PICKLE.
    # `pg.Dict` is a subclass of `dict`, therefore, when pickle loads a dict,
    # it will try to set its items directly by calling `__setitem__` without
    # calling `pg.Dict.__init__` at the first place. As a result, an error will
    # raise, which complains about that an attribute set up during `__init__` is
    # not available. A mitigation to this issue is to detect such calls in
    # `__setitem__` as the follows, and simply do nothing, which will give a
    # chance to `pg.Dict.__getstate__` to deal with the restoration logic as
    # an object (instead of a dict).
    if not hasattr(self, '_sym_parent'):
      return

    if base.treats_as_sealed(self):
      raise base.WritePermissionError(
          self._error_message('Cannot modify field of a sealed Dict.'))

    if not base.writtable_via_accessors(self):
      raise base.WritePermissionError(
          self._error_message(
              'Cannot modify Dict field by attribute or key while '
              'accessor_writable is set to False. '
              'Use \'rebind\' method instead.'))

    update = self._set_item_without_permission_check(key, value)
    if flags.is_change_notification_enabled() and update:
      self._notify_field_updates([update])

  def __setattr__(self, name: str, value: Any) -> None:
    """Set attribute of this Dict.

    NOTE(daiyip): When setting attributes, public attributes (not started with
    '_') are set as dict fields, while private attributes (started with '_') are
    set on the object instance.

    Args:
      name: Name of attribute.
      value: Value of attribute.
    """
    if name.startswith('_'):
      super().__setattr__(name, value)
    else:
      self[name] = value

  def __delitem__(self, name: str) -> None:
    """Delete a key from the Dict.

    This is used to delete a key which resolves to a pg.typing.NonConstKey.

    Args:
      name: Key to delete.

    Raises:
      WritePermissionError: When Dict is sealed.
      KeyError: When key is not a NonConstKey.
    """
    if base.treats_as_sealed(self):
      raise base.WritePermissionError('Cannot del item from a sealed Dict.')

    if not base.writtable_via_accessors(self):
      raise base.WritePermissionError(
          self._error_message('Cannot del Dict field by attribute or key while '
                              'accessor_writable is set to False. '
                              'Use \'rebind\' method instead.'))

    if name not in self:
      raise KeyError(
          self._error_message(f'Key does not exist in Dict: {name!r}.'))

    update = self._set_item_without_permission_check(
        name, pg_typing.MISSING_VALUE)
    if flags.is_change_notification_enabled() and update:
      self._notify_field_updates([update])

  def __delattr__(self, name: str) -> None:
    """Delete an attribute."""
    del self[name]

  def __getattr__(self, name: str) -> Any:
    """Get attribute that is not defined as property."""
    if name in self:
      return self.sym_inferred(name)
    raise AttributeError(
        f'Attribute \'{name}\' does not exist in {self.__class__!r}.')

  def __iter__(self):
    """Iterate keys in field declaration order."""
    return self.sym_keys()

  def keys(self) -> Iterator[str]:  # pytype: disable=signature-mismatch
    """Returns an iterator of keys in current dict."""
    return self.sym_keys()

  def items(self) -> Iterator[Tuple[str, Any]]:  # pytype: disable=signature-mismatch
    """Returns an iterator of (key, value) items in current dict."""
    return self.sym_items()

  def values(self) -> Iterator[Any]:  # pytype: disable=signature-mismatch
    """Returns an iterator of values in current dict.."""
    return self.sym_values()

  def copy(self) -> 'Dict':
    """Overridden copy using symbolic copy."""
    return self.sym_clone(deep=False)

  def pop(
      self, key: Any, default: Any = base.RAISE_IF_NOT_FOUND  # pylint: disable=protected-access
  ) -> Any:
    """Pops a key from current dict."""
    if key in self:
      value = self[key]
      with flags.allow_writable_accessors(True):
        del self[key]
      return value if value != pg_typing.MISSING_VALUE else default
    if default is base.RAISE_IF_NOT_FOUND:
      raise KeyError(key)
    return default

  def popitem(self) -> Tuple[str, Any]:
    if self._value_spec is not None:
      raise ValueError(
          '\'popitem\' cannot be performed on a Dict with value spec.')
    if base.treats_as_sealed(self):
      raise base.WritePermissionError('Cannot pop item from a sealed Dict.')
    return super().popitem()

  def clear(self) -> None:
    """Removes all the keys in current dict."""
    if base.treats_as_sealed(self):
      raise base.WritePermissionError('Cannot clear a sealed Dict.')
    value_spec = self._value_spec
    self._value_spec = None
    super().clear()

    if value_spec:
      self.use_value_spec(value_spec, self._allow_partial)

  def setdefault(self, key: str, default: Any = None) -> Any:
    """Sets default as the value to key if not present."""
    value = pg_typing.MISSING_VALUE
    if key in self:
      value = self.sym_getattr(key)
    if value == pg_typing.MISSING_VALUE:
      self[key] = default
      value = default
    return value

  def update(self,
             other: Union[
                 None,
                 typing.Dict[str, Any],
                 Iterable[Tuple[str, Any]]] = None,
             **kwargs) -> None:  # pytype: disable=signature-mismatch
    """Update Dict with the same semantic as update on standard dict."""
    updates = dict(other) if other else {}
    updates.update(kwargs)
    self.rebind(
        updates, raise_on_no_change=False, skip_notification=True)

  def sym_jsonify(
      self,
      hide_frozen: bool = True,
      hide_default_values: bool = False,
      exclude_keys: Optional[Sequence[str]] = None,
      use_inferred: bool = False,
      **kwargs) -> object_utils.JSONValueType:
    """Converts current object to a dict with plain Python objects."""
    exclude_keys = set(exclude_keys or [])
    if self._value_spec and self._value_spec.schema:
      json_repr = dict()
      matched_keys, _ = self._value_spec.schema.resolve(self.keys())  # pytype: disable=attribute-error
      for key_spec, keys in matched_keys.items():
        # NOTE(daiyip): The key values of frozen field can safely be excluded
        # since they will be the same for a class.
        field = self._value_spec.schema[key_spec]
        if hide_frozen and field.frozen:
          continue
        for key in keys:
          if key not in exclude_keys:
            value = self.sym_getattr(key)
            if use_inferred and isinstance(value, base.Inferential):
              value = self.sym_inferred(key, default=value)
            if pg_typing.MISSING_VALUE == value:
              continue
            if hide_default_values and base.eq(value, field.default_value):
              continue
            json_repr[key] = base.to_json(
                value,
                hide_frozen=hide_frozen,
                hide_default_values=hide_default_values,
                use_inferred=use_inferred,
                **kwargs)
      return json_repr
    else:
      return {
          k: base.to_json(
              self.sym_inferred(k, default=v) if (
                  use_inferred and isinstance(v, base.Inferential)) else v,
              hide_frozen=hide_frozen,
              hide_default_values=hide_default_values,
              use_inferred=use_inferred,
              **kwargs)
          for k, v in self.sym_items()
          if k not in exclude_keys
      }

  def custom_apply(
      self,
      path: object_utils.KeyPath,
      value_spec: pg_typing.ValueSpec,
      allow_partial: bool,
      child_transform: Optional[
          Callable[[object_utils.KeyPath, pg_typing.Field, Any], Any]] = None
  ) -> Tuple[bool, 'Dict']:
    """Implement pg.typing.CustomTyping interface.

    Args:
      path: KeyPath of current object.
      value_spec: Origin value spec of the field.
      allow_partial: Whether allow partial object to be created.
      child_transform: Function to transform child node values in dict_obj into
        their final values. Transform function is called on leaf nodes first,
        then on their containers, recursively.

    Returns:
      A tuple (proceed_with_standard_apply, transformed value)
    """
    proceed_with_standard_apply = True
    if self._value_spec:
      if value_spec and not value_spec.is_compatible(self._value_spec):
        raise ValueError(
            object_utils.message_on_path(
                f'Dict (spec={self._value_spec!r}) cannot be assigned to an '
                f'incompatible field (spec={value_spec!r}).', path))
      if self._allow_partial == allow_partial:
        proceed_with_standard_apply = False
      else:
        self._allow_partial = allow_partial
    elif isinstance(value_spec, pg_typing.Dict):
      self._value_spec = value_spec
    return (proceed_with_standard_apply, self)

  def format(
      self,
      compact: bool = False,
      verbose: bool = True,
      root_indent: int = 0,
      *,
      python_format: bool = False,
      hide_frozen: bool = True,
      hide_default_values: bool = False,
      hide_missing_values: bool = False,
      include_keys: Optional[Set[str]] = None,
      exclude_keys: Optional[Set[str]] = None,
      use_inferred: bool = False,
      cls_name: Optional[str] = None,
      bracket_type: object_utils.BracketType = object_utils.BracketType.CURLY,
      key_as_attribute: bool = False,
      extra_blankline_for_field_docstr: bool = False,
      **kwargs,
  ) -> str:
    """Formats this Dict."""
    cls_name = cls_name or ''
    exclude_keys = exclude_keys or set()
    def _indent(text, indent):
      return ' ' * 2 * indent + text

    def _should_include_key(key):
      if include_keys:
        return key in include_keys
      return key not in exclude_keys

    field_list = []
    if self._value_spec and self._value_spec.schema:
      matched_keys, unmatched = self._value_spec.schema.resolve(self.keys())  # pytype: disable=attribute-error
      assert not unmatched
      for key_spec, keys in matched_keys.items():
        for key in keys:
          if _should_include_key(key):
            field = self._value_spec.schema[key_spec]
            if hide_frozen and field.frozen:
              continue
            v = self.sym_getattr(key)
            if use_inferred and isinstance(v, base.Inferential):
              v = self.sym_inferred(key, default=v)
            if pg_typing.MISSING_VALUE == v:
              if hide_missing_values:
                continue
            elif hide_default_values and base.eq(v, field.default_value):
              continue
            field_list.append((field, key, v))
    else:
      for k, v in self.sym_items():
        if _should_include_key(k):
          if use_inferred and isinstance(v, base.Inferential):
            v = self.sym_inferred(k, default=v)
          field_list.append((None, k, v))

    open_bracket, close_bracket = object_utils.bracket_chars(bracket_type)
    if not field_list:
      return f'{cls_name}{open_bracket}{close_bracket}'

    if compact:
      s = [f'{cls_name}{open_bracket}']
      kv_strs = []
      for _, k, v in field_list:
        v_str = object_utils.format(
            v,
            compact,
            verbose,
            root_indent + 1,
            hide_frozen=hide_frozen,
            hide_default_values=hide_default_values,
            hide_missing_values=hide_missing_values,
            python_format=python_format,
            use_inferred=use_inferred,
            extra_blankline_for_field_docstr=extra_blankline_for_field_docstr,
            **kwargs)
        if not python_format or key_as_attribute:
          kv_strs.append(f'{k}={v_str}')
        else:
          kv_strs.append(f'\'{k}\': {v_str}')

      s.append(', '.join(kv_strs))
      s.append(close_bracket)
    else:
      s = [f'{cls_name}{open_bracket}\n']
      for i, (f, k, v) in enumerate(field_list):
        if i != 0:
          s.append(',\n')

        if verbose and f and typing.cast(pg_typing.Field, f).description:
          if i != 0 and extra_blankline_for_field_docstr:
            s.append('\n')
          description = typing.cast(pg_typing.Field, f).description
          for line in description.split('\n'):
            s.append(_indent(f'# {line}\n', root_indent + 1))
        v_str = object_utils.format(
            v,
            compact,
            verbose,
            root_indent + 1,
            hide_frozen=hide_frozen,
            hide_default_values=hide_default_values,
            hide_missing_values=hide_missing_values,
            python_format=python_format,
            use_inferred=use_inferred,
            extra_blankline_for_field_docstr=extra_blankline_for_field_docstr,
            **kwargs)
        if not python_format:
          # Format in PyGlove's format (default).
          s.append(_indent(f'{k} = {v_str}', root_indent + 1))
        elif key_as_attribute:
          # Format `pg.Objects` under Python format.
          s.append(_indent(f'{k}={v_str}', root_indent + 1))
        else:
          # Format regular `pg.Dict` under Python format.
          s.append(_indent(f'\'{k}\': {v_str}', root_indent + 1))
      s.append('\n')
      s.append(_indent(close_bracket, root_indent))
    return ''.join(s)

  def __repr__(self) -> str:
    """Operator repr()."""
    return base.Symbolic.__repr__(self)

  def __eq__(self, other: Any) -> bool:
    """Operator ==."""
    if isinstance(other, dict):
      return dict.__eq__(self, other)
    return False

  def __ne__(self, other: Any) -> bool:
    """Operator !=."""
    return not self.__eq__(other)

  def __hash__(self) -> int:
    """Overridden hashing function using symbolic hash."""
    return self.sym_hash()


base.Symbolic.DictType = Dict



# Copyright 2024 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Progress bar control."""

import functools
from typing import Annotated, List, Optional, Union

from pyglove.core import object_utils
from pyglove.core.symbolic import object as pg_object
# pylint: disable=g-importing-member
from pyglove.core.views.html.base import Html
from pyglove.core.views.html.controls.base import HtmlControl
from pyglove.core.views.html.controls.label import Label
# pylint: enable=g-importing-member


@pg_object.use_init_args(
    ['name', 'value', 'id', 'css_classes', 'styles']
)
class SubProgress(HtmlControl):
  """A sub progress bar control."""

  name: Annotated[
      str, 'The name of the sub progress bar.'
  ]

  value: Annotated[
      int, 'The value of the sub progress bar.'
  ] = 0

  interactive = True

  def _on_parent_change(self, *args, **kwargs):
    super()._on_parent_change(*args, **kwargs)  # pytype: disable=attribute-error
    self.__dict__.pop('parent', None)

  @functools.cached_property
  def parent(self) -> Optional['ProgressBar']:
    """Returns the parent progress bar."""
    return self.sym_ancestor(
        lambda x: isinstance(x, ProgressBar)
    )

  @property
  def total(self) -> Optional[int]:
    """Returns the total number of the sub progress bar."""
    assert self.parent is not None
    return self.parent.total

  @property
  def width(self) -> Optional[str]:
    """Returns the width of the sub progress bar."""
    if self.total is None:
      return None
    return f'{self.value / self.total:.0%}'

  def _to_html(self, **kwargs) -> Html:
    styles = self.styles.copy()
    styles.update(width=self.width)
    return Html.element(
        'div',
        [],
        id=self.element_id(),
        styles=styles,
        css_classes=[
            'sub-progress',
            object_utils.camel_to_snake(self.name, '-')
        ] + self.css_classes,
    )

  def increment(self, delta: int = 1):
    """Increments the value of the sub progress bar."""
    self.update(self.value + delta)

  def update(self, value: Optional[int] = None):
    if value is not None:
      self.rebind(
          value=value, skip_notification=True, raise_on_no_change=False
      )
    self._update_style(dict(width=self.width))
    self.parent.update()


@pg_object.use_init_args(
    ['subprogresses', 'total', 'id', 'css_classes', 'styles']
)
class ProgressBar(HtmlControl):
  """A progress bar control."""

  subprogresses: Annotated[
      List[SubProgress],
      'The sub progress bars of the progress bar.'
  ]

  total: Annotated[
      Optional[int],
      (
          'The total number of steps represented by the progress bar.'
          'If None, the progress steps are not determined.'
      )
  ] = None

  interactive = True

  def _on_bound(self):
    super()._on_bound()
    self._progress_label = Label(
        text=self._progress_text(),
        tooltip=self._progress_tooltip(),
        css_classes=['progress-label'],
        interactive=True,
    )

  def _progress_text(self) -> Union[str, Html]:
    completed = sum(s.value for s in self.subprogresses)
    if self.total is None:
      return 'n/a'
    complete_rate = completed / self.total
    return f'{complete_rate: .1%} ({completed}/{self.total})'

  def _progress_tooltip(self) -> str:
    if self.total is None:
      return 'Not started'
    assert self.total > 0
    return '\n'.join([
        f'{s.name}: {s.value / self.total:.1%} ({s.value}/{self.total})'
        for s in self.subprogresses
    ])

  def _to_html(self, **kwargs) -> Html:
    return Html.element(
        'div',
        [
            Html.element('div', self.subprogresses, css_classes=['shade']),
            self._progress_label,
        ],
        css_classes=['progress-bar']
    ).add_style(
        """
        .progress-bar {
            display: inline-block;
        }
        .progress-bar .shade {
          position: relative;
          background-color: #EEE;
          display: inline-flex;
          margin: 10px;
          width: 100px; 
          height: 5px;
          border-radius: 5px; 
          overflow: hidden;
        }
        .progress-bar .sub-progress {
          background-color: dodgerblue;
          height: 10px;
          width: 0%;
          border-radius: 0px;
        }
        """
    )

  def update(self, total: Optional[int] = None):
    if total is not None:
      assert self.total is None
      self._sync_members(total=total)
    self._progress_label.update(
        self._progress_text(), tooltip=self._progress_tooltip()
    )

  def __getitem__(self, name: str) -> SubProgress:
    for sub in self.subprogresses:
      if sub.name == name:
        return sub
    raise KeyError(f'Sub progress bar {name!r} not found.')



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tests for pyglove.object_utils.value_location."""

import unittest
from pyglove.core.object_utils import formatting
from pyglove.core.object_utils import value_location


KeyPath = value_location.KeyPath
KeyPathSet = value_location.KeyPathSet


class KeyPathTest(unittest.TestCase):
  """Tests for class KeyPath."""

  def test_basics(self):
    # Root element.
    r = KeyPath()
    self.assertTrue(r.is_root)
    self.assertFalse(r)
    self.assertEqual(r, '')
    self.assertEqual(r.depth, 0)
    self.assertEqual(len(r), 0)
    self.assertEqual(r.path, '')
    with self.assertRaisesRegex(
        KeyError, 'Parent of a root KeyPath does not exist.'):
      _ = r.parent

    with self.assertRaisesRegex(
        KeyError, 'Key of root KeyPath does not exist.'):
      _ = r.key

    # 1-level deep.
    a = KeyPath('a')
    self.assertFalse(a.is_root)
    self.assertEqual(a.key, 'a')
    self.assertEqual(a.path, 'a')
    self.assertEqual(a, 'a')  # Relative path compare.
    self.assertNotEqual(a, '')
    self.assertEqual(a.depth, 1)
    self.assertEqual(len(a), 1)
    self.assertEqual(a.parent, r)

    a2 = KeyPath(0)
    self.assertFalse(a2.is_root)
    self.assertEqual(a2.key, 0)
    with formatting.str_format(markdown=True):
      self.assertEqual(str(a2), '[0]')
    with formatting.repr_format(markdown=True):
      self.assertEqual(repr(a2), '[0]')
    self.assertEqual(a2.path, '[0]')
    self.assertEqual(a2, '[0]')
    self.assertNotEqual(a2, '')
    self.assertEqual(a2.depth, 1)
    self.assertEqual(len(a2), 1)
    self.assertEqual(a2.parent, r)

    a3 = KeyPath('x.y')
    self.assertFalse(a3.is_root)
    self.assertEqual(a3.key, 'x.y')
    self.assertEqual(a3.path, '[x.y]')
    self.assertEqual(a3.path_str(False), 'x.y')
    self.assertEqual(a3, '[x.y]')  # Relative path compare.
    self.assertNotEqual(a3, 'x.y')
    self.assertEqual(a3.depth, 1)
    self.assertEqual(len(a3), 1)
    self.assertEqual(a3.parent, r)

    # Multiple levels.
    b = KeyPath([1, 'b'])
    self.assertEqual(b, '[1].b')
    self.assertEqual(b.path, '[1].b')
    self.assertNotEqual(a, b)
    self.assertEqual(len(b), 2)
    self.assertEqual(b.parent, '[1]')

    c = KeyPath('c', b)
    self.assertEqual(c.key, 'c')
    self.assertEqual(c, '[1].b.c')
    self.assertEqual(c.keys, [1, 'b', 'c'])
    self.assertEqual(len(c), 3)
    self.assertEqual(c.parent, b)

    d = KeyPath(['d', 0], c)
    self.assertEqual(d.key, 0)
    self.assertEqual(d, '[1].b.c.d[0]')
    self.assertEqual(d.keys, [1, 'b', 'c', 'd', 0])
    self.assertEqual(d.parent, '[1].b.c.d')
    self.assertEqual(len(d), 5)

    d2 = KeyPath(('d', 0), c)
    self.assertEqual(d, d2)

  def test_complex_key_type(self):

    class A:

      def __init__(self, text):
        self._text = text

      def __str__(self):
        return f'A({self._text})'

    p = KeyPath([A('a'), A('b'), 'c'])
    self.assertEqual(p.path, '[A(a)][A(b)].c')

    # Key may have '.' in their string form.
    p = KeyPath([A('a.*'), A('$b')])
    self.assertEqual(p.path, '[A(a.*)][A($b)]')

    # NOTE: We cannot really parse KeyPath with complex types.

    class B(value_location.StrKey):
      """Class that implements StrKey will be treated as string key."""

      def __init__(self, text):
        self._text = text

      def __str__(self):
        return f'B({self._text})'

    p = KeyPath([B('a'), B('b'), 'c'])
    self.assertEqual(p.path, 'B(a).B(b).c')

  def test_parse(self):
    """Test KeyPath.parse method."""
    self.assertEqual(KeyPath.parse('a').keys, ['a'])
    self.assertEqual(len(KeyPath.parse('')), 0)

    self.assertEqual(KeyPath.parse('a').keys, ['a'])
    self.assertEqual(KeyPath.parse('[a ]').keys, ['a '])
    self.assertEqual(KeyPath.parse('[0].a').keys, [0, 'a'])
    self.assertEqual(
        KeyPath.parse('[0][1].a').keys, [0, 1, 'a'])
    self.assertEqual(
        KeyPath.parse('a.b[1].c').keys, ['a', 'b', 1, 'c'])
    self.assertEqual(
        KeyPath.parse('a[x[0]].b[y.z].c').keys,
        ['a', 'x[0]', 'b', 'y.z', 'c'])

    with self.assertRaisesRegex(
        ValueError, '\'path_str\' must be a string type.'):
      KeyPath.parse(0)

    with self.assertRaisesRegex(
        ValueError,
        'KeyPath parse failed: unmatched open bracket at position 0'):
      KeyPath.parse('[0')

    with self.assertRaisesRegex(
        ValueError,
        'KeyPath parse failed: unmatched open bracket at position 0'):
      KeyPath.parse('[[0]')

    with self.assertRaisesRegex(
        ValueError,
        'KeyPath parse failed: unmatched close bracket at position 3'):
      KeyPath.parse('[0]]')

  def test_from_value(self):
    """Test KeyPath.from_value."""
    self.assertEqual(
        KeyPath.from_value('x.y'),
        KeyPath(['x', 'y']))

    self.assertEqual(
        KeyPath.from_value(1),
        KeyPath([1]))

    path = KeyPath(['x'])
    self.assertIs(
        KeyPath.from_value(path),
        path)

    with self.assertRaisesRegex(
        ValueError, '.* is not a valid KeyPath equivalence'):
      KeyPath.from_value(0.1)

  def test_arithmetics(self):
    """Test KeyPath arithmetics."""

    # Test operator +.
    self.assertEqual(KeyPath('a') + 'b.c', 'a.b.c')
    self.assertEqual(KeyPath('a') + '[0].b', 'a[0].b')
    self.assertEqual(KeyPath('a') + None, 'a')
    self.assertEqual(KeyPath('a') + 1, 'a[1]')
    self.assertEqual(
        KeyPath('a') + KeyPath('b'), 'a.b')
    self.assertEqual(KeyPath.parse('a.b') + 1.0, 'a.b[1.0]')

    # Test operator -.
    self.assertEqual(
        KeyPath('a') - KeyPath('a'), '')
    self.assertEqual(KeyPath('a') - 'a', '')
    self.assertEqual(KeyPath('a') - '', 'a')
    self.assertEqual(KeyPath('a') - None, 'a')
    self.assertEqual(
        KeyPath('a') - KeyPath(), 'a')
    self.assertEqual(KeyPath.parse('a.b.c.d') - 'a.b', 'c.d')
    self.assertEqual(KeyPath.parse('[0].a') - 0, 'a')

    with self.assertRaisesRegex(
        ValueError, 'KeyPath subtraction failed: .* are in different subtree.'):
      _ = KeyPath('a') - 'b'

    with self.assertRaisesRegex(
        ValueError, 'KeyPath subtraction failed: .* are in different subtree.'):
      _ = KeyPath.parse('a.b') - 'a.c'

    with self.assertRaisesRegex(
        ValueError, 'KeyPath subtraction failed: .* are in different subtree.'):
      _ = KeyPath.parse('a[0]') - 'a[1]'

    with self.assertRaisesRegex(
        ValueError, 'KeyPath subtraction failed: .* is an ancestor'):
      _ = KeyPath.parse('a.b') - 'a.b.c'

    with self.assertRaisesRegex(TypeError, 'Cannot subtract KeyPath'):
      _ = KeyPath.parse('a.b') - 1.0

  def test_is_relative_to(self):
    self.assertTrue(
        KeyPath.parse('a.b.c').is_relative_to(
            KeyPath())
    )
    self.assertTrue(
        KeyPath.parse('a.b.c').is_relative_to(
            KeyPath.parse('a.b'))
    )
    self.assertTrue(
        KeyPath.parse('a.b.c').is_relative_to(
            KeyPath.parse('a.b.c'))
    )
    self.assertFalse(
        KeyPath.parse('a.b').is_relative_to(
            KeyPath.parse('a.b.c'))
    )
    self.assertFalse(
        KeyPath.parse('a.b.d').is_relative_to(
            KeyPath.parse('a.b.c'))
    )

  def test_hash(self):
    self.assertIn(KeyPath.parse('a.b.c'), {'a.b.c': 1})
    self.assertNotIn(KeyPath.parse('a.b.c'), {'a.b': 1})

  def test_comparison(self):
    keypath = KeyPath.parse
    # Equality should only hold true for KeyPaths that are identical.
    self.assertEqual(
        KeyPath(), KeyPath.parse(''))
    self.assertEqual(keypath('a[1][2].b[3][4]'), keypath('a[1][2].b[3][4]'))
    self.assertNotEqual(keypath('a[1][2].b[3][4]'), keypath('a[1][2].a[3][4]'))
    self.assertNotEqual(keypath('a[1][2].b[3][4]'), keypath('a[1][2].b[4][4]'))
    # Earlier keys in the path should be prioritized over later ones.
    self.assertLess(KeyPath(), 'a')
    self.assertLess(KeyPath(), keypath('a'))
    self.assertLess(keypath('a'), keypath('a.a'))
    self.assertLess(keypath('a.a'), keypath('a.b'))
    self.assertGreater(keypath('a'), KeyPath())
    self.assertGreater(keypath('a[1].b'), keypath('a[1].a'))
    self.assertGreater(keypath('a[1].a'), keypath('a[1]'))
    # Numbers should be compared numerically - not lexicographically.
    self.assertLessEqual(keypath('a[1]'), keypath('a[2]'))
    self.assertLessEqual(keypath('a[2]'), keypath('a[10]'))
    self.assertGreaterEqual(keypath('a[10]'), keypath('a[2]'))
    self.assertGreaterEqual(keypath('a[2]'), keypath('a[1]'))
    # It should be possible to compare numeric keys with string keys.
    self.assertLess(keypath('a[1]'), keypath('a.b'))
    self.assertGreater(keypath('a.b'), keypath('a[1]'))
    self.assertLessEqual(keypath('a[1]'), keypath('a.b'))
    self.assertGreaterEqual(keypath('a.b'), keypath('a[1]'))

    class CustomKey(value_location.StrKey):

      def __init__(self, text):
        self.text = text

      def __lt__(self, other):
        if isinstance(other, CustomKey):
          return self.text < other.text
        return False

    self.assertLess(
        KeyPath([CustomKey('a'), 'b']),
        KeyPath([CustomKey('b'), 'b']))

    with self.assertRaisesRegex(
        TypeError, 'Comparison is not supported between instances'):
      _ = KeyPath() < 1

  def test_query(self):

    def query_shall_succeed(path_str, obj, expected_value, use_inferred=False):
      self.assertEqual(
          KeyPath.parse(path_str).query(obj, use_inferred),
          expected_value)

    def query_shall_fail(path_str,
                         obj,
                         error='Cannot query sub-key .* of object .*'):
      with self.assertRaisesRegex(KeyError, error):
        KeyPath.parse(path_str).query(obj)

    def get_shall_succeed(path_str, obj, default, expected_value):
      self.assertEqual(
          KeyPath.parse(path_str).get(obj, default),
          expected_value)

    def assert_exists(path_str, obj, should_exists):
      self.assertEqual(
          KeyPath.parse(path_str).exists(obj), should_exists)

    # Query at root level.
    query_shall_succeed('', 1, 1)
    query_shall_succeed('', None, None)
    query_shall_succeed('', [1, 2], [1, 2])
    query_shall_succeed('', {'a': 'foo'}, {'a': 'foo'})

    # Query simple types with more than 1 depth.
    query_shall_fail('a', 1)
    query_shall_fail('[0]', None)

    # Query complex types.
    class Foo:
      """Custom object."""

      def __init__(self, values):
        self._values = values

      def __getitem__(self, key):
        return self._values[key]

      # NOTE(daiyip): __len__ and __contains__ is intentional omitted.

    class Bar:
      """Custom object with sym_getattr."""

      def __init__(self, **kwargs):
        self._map = kwargs

      def sym_hasattr(self, name):
        return name in self._map

      def sym_getattr(self, name):
        return self._map[name]

      def sym_inferred(self, key):
        return Bar(z=self._map[key])

      def __contains__(self, key):
        return key in self._map

      def __eq__(self, other):
        return self._map == other._map

      def __ne__(self, other):
        return not self.__eq__(other)

    src = {'a': [{'c': 'foo'},
                 {'d': [1, 2]}],
           'b': True,
           'e': Foo([1, 2, 3]),
           'f': Bar(x=0, y=1)}

    query_shall_succeed('', src, src)
    query_shall_succeed('a', src, src['a'])
    query_shall_succeed('a[0]', src, src['a'][0])
    query_shall_succeed('a[0].c', src, src['a'][0]['c'])
    query_shall_succeed('a[1].d[1]', src, src['a'][1]['d'][1])
    query_shall_succeed('b', src, src['b'])
    query_shall_succeed('f.x', src, 0)
    query_shall_succeed('f.x.z', src, Bar(z=0), use_inferred=True)

    query_shall_fail('c', src, 'Path .* does not exist: key .* is absent')
    query_shall_fail('a.c', src, 'Path .* does not exist: key .* is absent')
    query_shall_fail('a[2]', src, 'Path .* does not exist: key .* is absent')
    query_shall_fail('a[1].e', src, 'Path .* does not exist: key .* is absent')
    query_shall_fail(
        'e[0]', src, 'Cannot query index .* on object .*: '
        '\'__len__\' does not exist')
    query_shall_fail(
        'e.f', src, 'Cannot query key .* on object .*: '
        '\'__contains__\' does not exist')
    query_shall_fail(
        'f.z', src, 'Path .* does not exist: key .* is absent')
    query_shall_fail('f.x.z', src, 'Cannot query sub-key .* does not exist')

    # Test get method.
    get_shall_succeed('', src, None, src)
    get_shall_succeed('a[1].d[1]', src, None, src['a'][1]['d'][1])
    get_shall_succeed('c', src, None, None)
    get_shall_succeed('b.c', src, 1, 1)

    # Test exists method.
    assert_exists('', src, True)
    assert_exists('a[1].d[1]', src, True)
    assert_exists('c', src, False)
    assert_exists('b.c', src, False)

  def test_message_on_path(self):
    self.assertEqual(value_location.message_on_path('hi.', None), 'hi.')
    self.assertEqual(
        value_location.message_on_path('hi.', KeyPath()),
        'hi. (path=)')
    self.assertEqual(
        value_location.message_on_path('hi.', KeyPath(['a'])),
        'hi. (path=a)')


class KeyPathSetTest(unittest.TestCase):
  """Tests for class KeyPathSet."""

  def test_empty_set(self):
    s1 = KeyPathSet()
    self.assertFalse(s1)
    self.assertNotIn('', s1)
    self.assertNotIn(KeyPath(), s1)
    self.assertNotIn('abc', s1)
    self.assertNotIn(1, s1)
    self.assertEqual(list(s1), [])
    self.assertIs(s1.subtree(KeyPath()), s1)
    self.assertFalse(s1.subtree('a.b.c'))
    self.assertEqual(s1, KeyPathSet())
    self.assertNotEqual(s1, 1)
    self.assertNotEqual(s1, KeyPathSet([1]))

  def test_add(self):
    s1 = KeyPathSet(
        ['a.b.c', 1, KeyPath([1, 'x']), 'a.b']
    )
    self.assertEqual(
        s1._trie,
        {
            1: {
                'x': {
                    '$': True,
                },
                '$': True
            },
            'a': {
                'b': {
                    'c': {
                        '$': True
                    },
                    '$': True,
                },
            }
        }
    )
    self.assertNotIn('', s1)
    self.assertNotIn('a', s1)
    self.assertIn(KeyPath(['a', 'b']), s1)
    self.assertIn('a.b.c', s1)
    self.assertIn(1, s1)
    self.assertIn('[1]', s1)
    self.assertIn('[1].x', s1)
    self.assertIn(KeyPath([1, 'x']), s1)

    self.assertTrue(s1.add(''))
    self.assertIn('', s1)
    self.assertFalse(s1.add('a.b.c'))

    # Test include_intermediate.
    s1 = KeyPathSet()
    self.assertTrue(s1.add('a.b.c', include_intermediate=True))
    self.assertIn('a', s1)
    self.assertIn('a.b', s1)
    self.assertIn('a.b.c', s1)

  def test_remove(self):
    s1 = KeyPathSet(
        ['a.b.c', 1, KeyPath([1, 'x']), 'a.b', 'c.d']
    )
    self.assertFalse(s1.remove('b'))
    self.assertFalse(s1.remove('c'))
    self.assertTrue(s1.remove('a.b.c'))
    self.assertTrue(s1.remove('a.b'))
    self.assertTrue(s1.remove(1))
    self.assertEqual(
        s1._trie,
        {
            1: {
                'x': {
                    '$': True,
                },
            },
            'c': {
                'd': {
                    '$': True,
                },
            },
        }
    )
    self.assertNotIn(1, s1)
    self.assertTrue(s1.has_prefix(1))

  def test_iter(self):
    self.assertEqual(
        list(KeyPathSet(['', 'a.b.c', 1, KeyPath([1, 'x']), 'a.b'])),
        [
            KeyPath(), KeyPath.parse('a.b.c'), KeyPath.parse('a.b'),
            KeyPath([1]), KeyPath([1, 'x'])
        ]
    )

  def test_has_prefix(self):
    s1 = KeyPathSet(['a.b.c', 1, KeyPath([1, 'x']), 'a.b'])
    self.assertTrue(s1.has_prefix('a'))
    self.assertTrue(s1.has_prefix('a.b'))
    self.assertTrue(s1.has_prefix('a.b.c'))
    self.assertTrue(s1.has_prefix(KeyPath(['a'])))
    self.assertTrue(s1.has_prefix(1))
    self.assertFalse(s1.has_prefix(2))
    self.assertFalse(s1.has_prefix('a.b.c.d'))

  def test_subpaths(self):
    s1 = KeyPathSet(['a.b.c', 1, KeyPath([1, 'x']), 'a.b'])
    self.assertIs(s1.subtree(''), s1)
    self.assertEqual(
        s1.subtree('a'), KeyPathSet(['b.c', 'b'])
    )
    self.assertEqual(s1.subtree(1), KeyPathSet(['', 'x']))
    self.assertEqual(
        s1.subtree(1), KeyPathSet(['', 'x'])
    )

  def test_clear(self):
    s1 = KeyPathSet(['a.b.c', 1, KeyPath([1, 'x']), 'a.b'])
    s1.clear()
    self.assertEqual(s1, KeyPathSet())

  def test_copy(self):
    s1 = KeyPathSet(['a.b.c', 1, KeyPath([1, 'x']), 'a.b'])
    s2 = s1.copy()
    self.assertIsNot(s1, s2)
    self.assertIsNot(s1._trie, s2._trie)
    self.assertIsNot(s1._trie['a'], s2._trie['a'])
    self.assertIsNot(s1._trie['a']['b'], s2._trie['a']['b'])

  def test_update(self):
    s1 = KeyPathSet(['a.b.c', 1, KeyPath([1, 'x']), 'a.b'])
    s1.update(KeyPathSet(['a.b.d', 'a.c', '']))
    self.assertEqual(
        s1._trie,
        {
            1: {
                'x': {
                    '$': True,
                },
                '$': True,
            },
            'a': {
                'b': {
                    'c': {
                        '$': True
                    },
                    'd': {
                        '$': True
                    },
                    '$': True,
                },
                'c': {
                    '$': True
                },
            },
            '$': True,
        }
    )

  def test_union(self):
    s1 = KeyPathSet(['a.b.c', 1, KeyPath([1, 'x']), 'a.b'])
    s2 = s1.union(KeyPathSet(['a.b.d', 'a.c', '']))
    self.assertEqual(
        list(s2),
        [
            KeyPath.parse('a.b.c'),
            KeyPath.parse('a.b'),
            KeyPath.parse('a.b.d'),
            KeyPath.parse('a.c'),
            KeyPath([1]),
            KeyPath([1, 'x']),
            KeyPath(),
        ]
    )
    self.assertIsNot(s2._trie['a'], s1._trie['a'])
    self.assertIsNot(s2._trie['a']['b'], s1._trie['a']['b'])
    self.assertIsNot(
        s2._trie['a']['b']['c'], s1._trie['a']['b']['c']
    )

  def test_difference(self):
    s1 = KeyPathSet(['a.b.c', 1, KeyPath([1, 'x']), 'a.b', ''])
    s2 = s1.difference(
        KeyPathSet(['a.b', 'a.c.b', '[1].x', ''])
    )
    self.assertEqual(
        s2._trie,
        {
            1: {
                '$': True
            },
            'a': {
                'b': {
                    'c': {
                        '$': True
                    }
                }
            }
        }
    )
    self.assertIsNot(s2._trie['a'], s1._trie['a'])
    self.assertIsNot(s2._trie['a']['b'], s1._trie['a']['b'])
    self.assertIsNot(s2._trie[1], s1._trie[1])

    s1.difference_update(KeyPathSet(['a.b', 'a.c.b', '[1].x', '']))
    self.assertEqual(list(s1), ['a.b.c', '[1]'])

  def test_intersection(self):
    s1 = KeyPathSet(['a.b.c', 1, KeyPath([1, 'x']), 'a.b', 'a.c.d', ''])
    s2 = s1.intersection(
        KeyPathSet(['a.b', 'a.b.d', 'a.c.b', '[1].x', ''])
    )
    self.assertEqual(
        s2._trie,
        {
            1: {
                'x': {
                    '$': True,
                },
            },
            'a': {
                'b': {
                    '$': True,
                },
            },
            '$': True,
        }
    )
    self.assertIsNot(s2._trie['a'], s1._trie['a'])
    self.assertIsNot(s2._trie['a']['b'], s1._trie['a']['b'])
    self.assertIsNot(s2._trie[1], s1._trie[1])

    s1.intersection_update(
        KeyPathSet(['a.b', 'a.c.b', '[1].x', ''])
    )
    self.assertEqual(list(s1), ['a.b', '[1].x', ''])

  def test_rebase(self):
    s1 = KeyPathSet(['x.y', 'y', 'y.z.w'])
    s1.rebase('a.b.')
    self.assertEqual(
        list(s1),
        [
            KeyPath.parse('a.b.x.y'),
            KeyPath.parse('a.b.y'),
            KeyPath.parse('a.b.y.z.w'),
        ]
    )

  def test_operator_add(self):
    self.assertEqual(
        KeyPathSet(['a.b.c', 'a.b']) + KeyPathSet(['a.b', '[1].a', '']),
        KeyPathSet(['a.b.c', 'a.b', '[1].a', ''])
    )
    self.assertEqual(
        KeyPath.parse('x[0]') + KeyPathSet(['a.b.c', 'a.b']),
        KeyPathSet(['x[0].a.b.c', 'x[0].a.b'])
    )

  def test_format(self):
    self.assertEqual(
        KeyPathSet(['a.b.c', 'a.b']).format(),
        'KeyPathSet([a.b.c, a.b])'
    )

  def test_from_value(self):
    """Test KeyPathSet.from_value."""
    s = KeyPathSet(['a.b.c'])
    self.assertIs(
        KeyPathSet.from_value(s), s
    )
    self.assertEqual(
        KeyPathSet.from_value(['a.b']),
        KeyPathSet(['a.b'])
    )
    self.assertEqual(
        KeyPathSet.from_value(['a.b'], include_intermediate=True),
        KeyPathSet(['a', 'a.b', ''])
    )
    with self.assertRaisesRegex(
        ValueError, 'Cannot convert .* to KeyPathSet'
    ):
      KeyPathSet.from_value(1)

if __name__ == '__main__':
  unittest.main()



# Copyright 2023 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import inspect
import unittest
from pyglove.core.object_utils import error_utils


class ErrorInfoTest(unittest.TestCase):
  """Tests for ErrorInfo."""

  def test_from_exception(self):

    def foo():
      return 1 / 0

    def bar():
      try:
        foo()
      except ZeroDivisionError as e:
        raise ValueError('Bad call to `foo`') from e

    error_info = None
    try:
      bar()
    except ValueError as e:
      error_info = error_utils.ErrorInfo.from_exception(e)
    self.assertIsNotNone(error_info)
    self.assertEqual(error_info.tag, 'ValueError.ZeroDivisionError')
    self.assertEqual(error_info.description, 'Bad call to `foo`')
    self.assertIn('Traceback (most recent call last)', error_info.stacktrace)

  def test_to_json(self):
    error_info = error_utils.ErrorInfo(
        tag='ValueError.ZeroDivisionError',
        description='Bad call to `foo`',
        stacktrace='Traceback (most recent call last)',
    )
    json_dict = error_info.to_json()
    error_info2 = error_utils.ErrorInfo.from_json(json_dict)
    self.assertIsNot(error_info2, error_info)
    self.assertEqual(error_info2, error_info)

  def test_format(self):
    error_info = error_utils.ErrorInfo(
        tag='ValueError.ZeroDivisionError',
        description='Bad call to `foo`',
        stacktrace='Traceback (most recent call last)',
    )
    self.assertEqual(
        error_info.format(compact=False),
        inspect.cleandoc(
            """
            ErrorInfo(
              tag='ValueError.ZeroDivisionError',
              description='Bad call to `foo`',
              stacktrace='Traceback (most recent call last)'
            )
            """
        )
    )


class CatchErrorsTest(unittest.TestCase):

  def assert_caught_error(self, func, errors_to_catch):
    with error_utils.catch_errors(errors_to_catch) as context:
      func()
    self.assertIsNotNone(context.error)

  def assert_propagate_error(self, func, errors_to_catch):

    with self.assertRaises(Exception):
      with error_utils.catch_errors(errors_to_catch):
        func()

  def test_catch_errors(self):
    def foo():
      raise ValueError('this is an error')

    self.assert_caught_error(foo, ValueError)
    self.assert_caught_error(foo, (ValueError,))
    self.assert_caught_error(foo, (Exception, 'ValueError'))
    self.assert_caught_error(foo, (KeyError, ValueError))
    self.assert_caught_error(foo, (ValueError, 'an error'))
    self.assert_caught_error(foo, (KeyError, (ValueError, 'an error'),))

    self.assert_propagate_error(foo, KeyError)
    self.assert_propagate_error(foo, (ValueError, '^an error'))
    self.assert_propagate_error(foo, (ValueError, 'something else'))

  def test_catch_errors_with_error_handler(self):
    errors = []
    def handle_error(error):
      errors.append(error)

    def foo():
      raise ValueError()

    with error_utils.catch_errors(ValueError, handle_error) as context:
      foo()
    self.assertEqual(errors, [context.error])

  def test_catch_errors_bad_inputs(self):
    with self.assertRaisesRegex(
        TypeError, 'Each error specification should be either .*'):
      with error_utils.catch_errors([(ValueError, 'abc', 'abc')]):
        pass

    with self.assertRaisesRegex(
        TypeError, 'Each error specification should be either .*'):
      with error_utils.catch_errors([(ValueError, 1)]):
        pass

    with self.assertRaisesRegex(
        TypeError, 'Exception contains non-except types'):
      with error_utils.catch_errors([str, ValueError]):
        pass


if __name__ == '__main__':
  unittest.main()



# Copyright 2022 The PyGlove Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Interface for tuning backend and backend factory."""

import abc
from typing import List, Optional, Sequence, Type

from pyglove.core import geno
from pyglove.core.tuning.early_stopping import EarlyStoppingPolicy
from pyglove.core.tuning.protocols import Feedback
from pyglove.core.tuning.protocols import Result


class Backend(metaclass=abc.ABCMeta):
  """Interface for the tuning backend."""

  @classmethod
  def create(cls,
             name: Optional[str],
             group: Optional[str],
             dna_spec: geno.DNASpec,
             algorithm: geno.DNAGenerator,
             metrics_to_optimize: Sequence[str],
             early_stopping_policy: Optional[EarlyStoppingPolicy] = None,
             num_examples: Optional[int] = None,
             **kwargs) -> 'Backend':
    """Create an instance of `Backend` based on `pg.sample` arguments.

    The default implementation is to pass through all the arguments to
    ``__init__`` for creating an instance of the backend. Users can override.

    Args:
      name: A string as a unique identifier for current sampling. Two separate
        calls to `pg.sample` with the same `name` (also the same algorithm) will
        share the same sampling queue, whose examples are proposed by the same
        search algorithm.
      group: An string or integer as the group ID of current process in
        distributed sampling, which will be used to group different workers into
        co-worker groups. Workers with the same group id will work on the same
        trial. On the contrary, workers in different groups will always be
        working with different trials. If not specified, each worker in current
        sampling will be in different groups. `group` is usually used in the
        outer loops of nested search, in order to allow workers to work on the
        same higher-order item.
      dna_spec: An `pg.DNASpec` object representing the search space.
      algorithm: The search algorithm that samples the search space.
      metrics_to_optimize: A sequence of string as the names of the metrics
        to be optimized by the algorithm, which is ['reward'] by default.
        When specified, it should have only 1 item for single-objective
        algorithm and can have multiple items for algorithms that support
        multi-objective optimization.
      early_stopping_policy: An optional early stopping policy for user to tell
        if incremental evaluation (which reports multiple measurements) on each
        example can be early short circuited.
        After each call to `feedback.add_measurement`, users can use method
        `feedback.should_stop_early` to check whether current example worth
        further evaluation or not.
      num_examples: An optional integer as the max number of examples to
        sample. If None, sample will return an iterator of infinite examples.
      **kwargs: Arguments passed to the `BackendFactory` subclass registered
        with the requested backend.

    Returns:
      A `pg.tuning.Backend` object.
    """
    return cls(     # pytype: disable=wrong-keyword-args
        name=name,
        group=group,
        dna_spec=dna_spec,
        algorithm=algorithm,
        metrics_to_optimize=metrics_to_optimize,
        early_stopping_policy=early_stopping_policy,
        num_examples=num_examples,
        **kwargs)

  @classmethod
  @abc.abstractmethod
  def poll_result(cls, name: str, **kwargs) -> Result:
    """Gets tuning result by a unique tuning identifier."""

  @abc.abstractmethod
  def next(self) -> Feedback:
    """Get the feedback object for the next sample."""


_backend_registry = dict()
_default_backend_name = 'in-memory'


def add_backend(backend_name: str):
  """Decorator to register a backend factory with name."""
  def _decorator(backend_cls):
    if not issubclass(backend_cls, Backend):
      raise TypeError(f'{backend_cls!r} is not a `pg.tuning.Backend` subclass.')
    _backend_registry[backend_name] = backend_cls
    return backend_cls
  return _decorator


def available_backends() -> List[str]:
  """Gets available backend names."""
  return list(_backend_registry.keys())


def set_default_backend(backend_name: str):
  """Sets the default tuning backend name."""
  if backend_name not in _backend_registry:
    raise ValueError(f'Backend {backend_name!r} does not exist.')
  global _default_backend_name
  _default_backend_name = backend_name


def default_backend() -> str:
  """Gets the default tuning backend name."""
  return _default_backend_name


def get_backend_cls(backend_name: str) -> Type[Backend]:
  """Get backend by name."""
  backend_name = backend_name or default_backend()
  if backend_name not in _backend_registry:
    raise ValueError(f'Backend {backend_name!r} does not exist.')
  return _backend_registry[backend_name]


def poll_result(
    name: str,
    backend: Optional[str] = None,
    **kwargs) -> Result:
  """Gets tuning result by name."""
  return get_backend_cls(backend).poll_result(name, **kwargs)
